# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
import pytest
import typer
from unittest.mock import MagicMock
from typer.testing import CliRunner

from dorsal.cli.adapter_app.export_cmd import export_adapter
from dorsal.cli.themes.palettes import DEFAULT_PALETTE
from dorsal.common.exceptions import DorsalError


cli_app = typer.Typer()


@cli_app.callback()
def main_callback(ctx: typer.Context):
    ctx.obj = {"palette": DEFAULT_PALETTE}


cli_app.command(name="export")(export_adapter)

runner = CliRunner()


@pytest.fixture
def mock_export_deps(mocker, mock_rich_console):
    """
    Mocks backend dependencies for the `export` command.
    """
    mock_error_console = MagicMock()
    mocker.patch("dorsal.common.cli.get_rich_console", return_value=mock_rich_console)
    mocker.patch("dorsal.common.cli.get_error_console", return_value=mock_error_console)

    mock_extract = mocker.patch("dorsal.cli.adapter_app.helpers.extract_records")
    mock_extract.return_value = [("MockSchema", {"text": "Extracted string"}, None)]

    mock_export_file = mocker.patch("dorsal.api.adapters.export_record_to_file")
    mock_export_file.return_value = "Mocked exported format content"

    mock_export_record = mocker.patch("dorsal.api.adapters.export_record")
    mock_export_record.return_value = "Mocked exported format content"

    return {
        "extract": mock_extract,
        "export_file": mock_export_file,
        "export_record": mock_export_record,
        "error_console": mock_error_console,
    }


def test_export_basic_success(mock_rich_console, mock_export_deps):
    """Tests a standard single-file export to stdout and auto-save via the API wrapper."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test_file.dorsal.json")
        test_file.write_text(json.dumps({"dummy": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "srt"])

        assert result.exit_code == 0, result.output

        mock_export_deps["export_file"].assert_called_once()
        kwargs = mock_export_deps["export_file"].call_args.kwargs
        assert kwargs["schema_id"] == "MockSchema"
        assert kwargs["target_format"] == "srt"
        assert kwargs["output_path"] == pathlib.Path.cwd() / "test_file.srt"

        printed_text = mock_rich_console.print.call_args.args[0]
        assert "Mocked exported format content" in str(printed_text)


def test_export_infer_format_from_output(mock_rich_console, mock_export_deps):
    """NEW TEST: Verifies that target_format can be completely omitted if inferred from --output."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test_file.dorsal.json")
        test_file.write_text(json.dumps({"dummy": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "--output", "my_custom_subs.vtt"])

        assert result.exit_code == 0

        kwargs = mock_export_deps["export_file"].call_args.kwargs
        assert kwargs["target_format"] == "vtt"
        assert kwargs["output_path"] == pathlib.Path("my_custom_subs.vtt")


def test_export_batch_success(mock_rich_console, mock_export_deps):
    """Tests batch processing which should render a table instead of raw text."""
    mock_export_deps["extract"].return_value = [
        ("MockSchema", {"text": "Rec 1"}, None),
        ("MockSchema", {"text": "Rec 2"}, None),
    ]

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test_file.dorsal.json")
        test_file.write_text(json.dumps([{"dummy": "data1"}, {"dummy": "data2"}]), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "md"])

        assert result.exit_code == 0

        assert mock_export_deps["export_file"].call_count == 2
        call_1_kwargs = mock_export_deps["export_file"].call_args_list[0].kwargs
        call_2_kwargs = mock_export_deps["export_file"].call_args_list[1].kwargs

        assert call_1_kwargs["output_path"].name == "test_file_1.md"
        assert call_2_kwargs["output_path"].name == "test_file_2.md"

        table_call_arg = mock_rich_console.print.call_args.args[0]
        assert type(table_call_arg).__name__ == "Table"
        assert "Export Results" in str(table_call_arg.title)


def test_export_invalid_json(mock_export_deps):
    """Tests handling of a malformed JSON input file."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("bad_file.json")
        test_file.write_text("{ bad json : ", encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "srt"])

        assert result.exit_code != 0
        error_msg = str(mock_export_deps["error_console"].print.call_args.args[0])
        assert "Invalid JSON file" in error_msg


def test_export_validation_error(mock_export_deps):
    """Tests handling of ValueError thrown by the record extractor."""
    mock_export_deps["extract"].side_effect = ValueError("Missing schema information")

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"bad": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "vtt"])

        assert result.exit_code != 0
        error_msg = str(mock_export_deps["error_console"].print.call_args.args[0])
        assert "Validation Error" in error_msg
        assert "Missing schema information" in error_msg


def test_export_missing_adapters_package(mock_export_deps):
    """Tests graceful failure when the optional dorsalhub-adapters is missing."""

    mock_export_deps["export_file"].side_effect = DorsalError(
        "Please pip install dorsalhub-adapters to enable exports."
    )

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"valid": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "srt"])

        assert result.exit_code != 0
        error_msg = str(mock_export_deps["error_console"].print.call_args.args[0])
        assert "dorsalhub-adapters" in error_msg


def test_export_adapter_runtime_error(mock_export_deps):
    """Tests handling when a specific adapter fails to process the record."""
    mock_export_deps["export_file"].side_effect = Exception("Format constraint violated")

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"valid": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "md"])

        assert result.exit_code != 0
        error_msg = str(mock_export_deps["error_console"].print.call_args.args[0])
        assert "Export Failed for schema" in error_msg
        assert "Format constraint violated" in error_msg


def test_export_no_save_flag(mock_export_deps):
    """Tests that the --no-save flag routes to export_record instead of export_record_to_file."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"dummy": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "srt", "--no-save"])

        assert result.exit_code == 0

        mock_export_deps["export_file"].assert_not_called()
        mock_export_deps["export_record"].assert_called_once()


def test_export_schema_override_passed_down(mock_export_deps):
    """Tests that --schema-id correctly passes the value to the extractor."""
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"dummy": "data"}), encoding="utf-8")

        runner.invoke(cli_app, ["export", str(test_file), "srt", "--schema-id", "CustomSchema"])

        mock_export_deps["extract"].assert_called_once_with({"dummy": "data"}, schema_override="CustomSchema")


def test_export_batch_with_adapter_error(mock_rich_console, mock_export_deps):
    """
    Tests batch processing where one record fails to export.
    This triggers the table rendering logic for item["status"] == "Error" and row.append("-").
    """
    mock_export_deps["extract"].return_value = [
        ("MockSchema", {"text": "Rec 1"}, None),
        ("MockSchema", {"text": "Rec 2"}, None),
    ]

    mock_export_deps["export_file"].side_effect = ["Mocked success content", Exception("Simulated batch failure")]

    with runner.isolated_filesystem():
        test_file = pathlib.Path("test_batch.dorsal.json")
        test_file.write_text(json.dumps([{"dummy": "data1"}, {"dummy": "data2"}]), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "md"])

        assert result.exit_code == 0
        assert mock_export_deps["export_file"].call_count == 2

        table_call_arg = mock_rich_console.print.call_args.args[0]
        assert type(table_call_arg).__name__ == "Table"


def test_export_with_orig_file_path(mock_export_deps):
    """
    Tests that if the record has an orig_file_path, it derives the output
    file name from that path's stem instead of the raw JSON's name.
    """
    mock_export_deps["extract"].return_value = [
        ("MockSchema", {"text": "Rec"}, "internal_dir/my_custom_image_name.png")
    ]

    with runner.isolated_filesystem():
        test_file = pathlib.Path("generic_wrapper.dorsal.json")
        test_file.write_text(json.dumps({"dummy": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file), "srt"])

        assert result.exit_code == 0

        kwargs = mock_export_deps["export_file"].call_args.kwargs
        assert kwargs["output_path"].name == "my_custom_image_name.srt"


def test_export_usage_error_no_format_or_output(mock_export_deps):
    """
    NEW TEST: Tests the usage error check if the user omits the format AND
    doesn't provide an --output flag that can be inferred.
    """
    with runner.isolated_filesystem():
        test_file = pathlib.Path("test.json")
        test_file.write_text(json.dumps({"valid": "data"}), encoding="utf-8")

        result = runner.invoke(cli_app, ["export", str(test_file)])

        assert result.exit_code != 0

        error_msg = str(mock_export_deps["error_console"].print.call_args.args[0])
        assert "Usage Error" in error_msg
        assert "target_format" in error_msg



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import json
from typing import Annotated, Optional, List

import typer
from rich.panel import Panel
from rich.text import Text

logger = logging.getLogger(__name__)


def _get_hashes(hash_list: Optional[List[str]], from_stdin: bool, json_output: bool, console) -> List[str]:
    all_hashes = set(hash_list) if hash_list else set()

    if from_stdin:
        if not sys.stdin.isatty():
            stdin_hashes = {line.strip() for line in sys.stdin if line.strip()}
            all_hashes.update(stdin_hashes)
        elif not json_output:
            console.print("[warning]--from-stdin flag was used, but no input was received.[/warning]")

    if not all_hashes:
        from dorsal.common.cli import exit_cli, EXIT_CODE_ERROR

        exit_cli(
            code=EXIT_CODE_ERROR,
            message="No file hashes provided. Use --hash or pipe input via --from-stdin.",
        )
    return list(all_hashes)


def remove_files(
    ctx: typer.Context,
    collection_id: Annotated[str, typer.Argument(help="The ID of the collection to remove files from.")],
    hash_list: Annotated[
        Optional[List[str]],
        typer.Option(
            "--hash",
            "-h",
            help="A file hash to remove. This option can be used multiple times.",
            show_default=False,
        ),
    ] = None,
    from_stdin: Annotated[
        bool,
        typer.Option(
            "--from-stdin",
            help="Read a list of hashes from standard input (one hash per line).",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option("--json", help="Output the results as a raw JSON object."),
    ] = False,
):
    """
    Remove one or more files from a remote collection by their hash.
    """
    from dorsal.api.collection import remove_files_from_collection
    from dorsal.common.cli import get_rich_console, exit_cli, EXIT_CODE_ERROR
    from dorsal.common.exceptions import AuthError, DorsalClientError, DorsalOfflineError

    console = get_rich_console()
    palette: dict[str, str] = ctx.obj["palette"]

    final_hash_list = _get_hashes(hash_list, from_stdin, json_output, console)

    try:
        if not json_output:
            with console.status(
                f"Removing {len(final_hash_list)} file(s) from collection '[bold]{collection_id}[/]'..."
            ):
                response = remove_files_from_collection(collection_id=collection_id, hashes=final_hash_list)
        else:
            response = remove_files_from_collection(collection_id=collection_id, hashes=final_hash_list)

    except ValueError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=str(e))
    except DorsalOfflineError:
        raise
    except AuthError:
        raise
    except DorsalClientError as e:
        exit_cli(code=EXIT_CODE_ERROR, message=f"API Error: {e.message}")
    except Exception as e:
        logger.exception("An unexpected error occurred in 'collection remove-files'")
        exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {e}")

    if json_output:
        console.print(response.model_dump_json(indent=2))
        exit_cli()

    output_text = Text()
    output_text.append("✅ Operation complete.\n", style=palette.get("success", "green"))
    output_text.append(f"Removed: {response.removed_count}\n")

    if response.not_found_count > 0:
        output_text.append(
            f"Not Found (ignored): {response.not_found_count}",
            style=palette.get("info", "dim"),
        )

    success_panel = Panel(
        output_text,
        title=f"[{palette.get('panel_title_success', 'bold green')}]Update Complete[/]",
        border_style=palette.get("panel_border_success", "green"),
        expand=False,
    )
    console.print(success_panel)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import pathlib
from typer.testing import CliRunner
from unittest.mock import MagicMock, ANY


from dorsal.cli import app
from dorsal.cli.collection_app import export_cmd
from dorsal.common.exceptions import DorsalClientError

runner = CliRunner()
COLLECTION_ID = "col_abc123"


@pytest.fixture
def mock_export_collection_cmd(mocker):
    """Mocks dependencies for the `collection export` command."""

    mock_export = mocker.patch("dorsal.api.collection.export_collection")

    mock_exports_dir = MagicMock(spec=pathlib.Path)

    default_save_dir = MagicMock(spec=pathlib.Path)
    final_path_obj = MagicMock(spec=pathlib.Path)

    mock_exports_dir.__truediv__.return_value = default_save_dir
    default_save_dir.__truediv__.return_value = final_path_obj

    expected_path_str = f"/fake/exports/{COLLECTION_ID}/export-{COLLECTION_ID}-20250101-120000.json.gz"
    final_path_obj.__str__.return_value = expected_path_str

    mocker.patch("dorsal.common.constants.CLI_EXPORTS_DIR", mock_exports_dir)

    # Mock time.strftime to create a predictable timestamp
    mocker.patch.object(export_cmd.time, "strftime", return_value="20250101-120000")

    return {
        "export_collection": mock_export,
        "exports_dir": mock_exports_dir,
    }


def test_export_collection_default_path(mock_rich_console, mock_export_collection_cmd):
    """Tests exporting with the default output path logic."""
    result = runner.invoke(app, ["collection", "export", COLLECTION_ID])

    assert result.exit_code == 0

    save_dir_mock = mock_export_collection_cmd["exports_dir"]
    save_dir_mock.mkdir.assert_called_once_with(parents=True, exist_ok=True)

    expected_file = save_dir_mock / f"export-{COLLECTION_ID}-20250101-120000.json.gz"

    mock_export_collection_cmd["export_collection"].assert_called_once_with(
        collection_id=COLLECTION_ID,
        output_path=expected_file,
        console=mock_rich_console,
        palette=ANY,
    )
    all_printed_text = "".join(str(call.args[0]) for call in mock_rich_console.print.call_args_list)
    assert "Export complete" in all_printed_text


def test_export_collection_with_output_dir(mock_export_collection_cmd, tmp_path):
    """Tests exporting to a user-specified directory."""
    result = runner.invoke(app, ["collection", "export", COLLECTION_ID, "--output-dir", str(tmp_path)])

    assert result.exit_code == 0

    expected_file = tmp_path / f"{COLLECTION_ID}-20250101-120000.json.gz"

    assert mock_export_collection_cmd["export_collection"].call_args.kwargs["output_path"] == expected_file


def test_export_collection_json_output(mock_rich_console, mock_export_collection_cmd):
    """Tests a successful export with --json output."""
    expected_filename = f"{COLLECTION_ID}-20250101-120000.json.gz"
    expected_full_path_str = f"/tmp/dorsal/exports/{expected_filename}"
    save_dir_mock = mock_export_collection_cmd["exports_dir"]
    output_file_mock = MagicMock(spec=pathlib.Path)
    save_dir_mock.__truediv__.return_value = output_file_mock
    output_file_mock.__str__.return_value = expected_full_path_str

    result = runner.invoke(app, ["collection", "export", COLLECTION_ID, "--json"])
    assert result.exit_code == 0

    mock_rich_console.print.assert_called_once()
    json_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_str)
    assert data["success"] is True
    assert data["collection_id"] == COLLECTION_ID
    assert data["output_path"] == expected_full_path_str

    assert expected_filename in data["output_path"]


def test_export_collection_api_error(mock_export_collection_cmd):
    """Tests graceful failure on a DorsalClientError from the backend."""
    mock_export_collection_cmd["export_collection"].side_effect = DorsalClientError("Collection is empty")

    result = runner.invoke(app, ["collection", "export", COLLECTION_ID])

    assert result.exit_code != 0
    assert "API Error: Collection is empty" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import datetime
import logging
import os
import time
from typing import (
    Iterator,
    Type,
    TYPE_CHECKING,
    cast,
    Iterable,
    Sequence,
    Any,
)

from rich.console import Console
from tqdm import tqdm

from dorsal.file.collection.base import _BaseFileCollection
from dorsal.client import DorsalClient
from dorsal.common.auth import is_offline_mode
from dorsal.common.environment import is_jupyter_environment
from dorsal.common.exceptions import (
    DorsalError,
    DorsalClientError,
    InvalidTagError,
    SyncConflictError,
)
from dorsal.common.constants import API_MAX_BATCH_SIZE
from dorsal.file.dorsal_file import LocalFile
from dorsal.file.metadata_reader import MetadataReader
from dorsal.file.permissions import is_permitted_public_media_type
from dorsal.session import get_shared_dorsal_client

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from dorsal.file.collection.remote import DorsalFileCollection
    from dorsal.file.validators.file_record import NewFileTag, FileRecord


def _get_source_paths(source_info: dict) -> list[str]:
    """Recursively extracts all source directory paths from a source_info dictionary."""
    paths = []
    if source_info.get("type") == "local" and source_info.get("path"):
        paths.append(source_info["path"])
    elif source_info.get("type") == "merged":
        for source in source_info.get("sources", []):
            paths.extend(_get_source_paths(source))
    return list(dict.fromkeys(paths))


class LocalFileCollection(_BaseFileCollection):
    """
    A high-level interface to create and manage a collection of local files.

    This class can be initialized from a directory path, which it will scan,
    or from a pre-existing list of LocalFile objects. It provides methods for
    pushing file metadata to DorsalHub, creating remote collections, and
    performing bulk operations like tagging.
    """

    files: Sequence[LocalFile]

    def __init__(
        self,
        source: str | list[LocalFile] | None = None,
        *,
        files: Sequence[LocalFile] | None = None,
        recursive: bool = False,
        client: DorsalClient | None = None,
        model_runner_pipeline: str | list[dict[str, Any]] | None = "default",
        file_class: Type[LocalFile] = LocalFile,
        source_info: dict | None = None,
        palette: dict | None = None,
        console: Console | None = None,
        use_cache: bool = True,
        overwrite_cache: bool = False,
        offline: bool = False,
        follow_symlinks: bool = True,
        lazy: bool = False,
    ):
        """
        Initializes the LocalFileCollection.

        The constructor can either scan a directory to build a new collection of
        files or wrap an existing list of LocalFile objects.

        Args:
            source (Union[str, list[LocalFile]]): A directory path to scan or a
                pre-populated list of LocalFile objects.
            recursive (bool): If scanning a directory, whether to include
                subdirectories. Defaults to False.
            client (DorsalClient | None): An optional pre-initialized DorsalClient
                instance to use for API operations. Defaults to None.
            model_runner_pipeline (Union[dict, str, None]): An optional custom
                model runner pipeline configuration. Defaults to None.
            file_class (Type[LocalFile]): The class to use when instantiating
                files from disk. Defaults to LocalFile.
            source_info (dict | None): Optional metadata about the source. Used
                internally when merging collections. Defaults to None.
            palette (dict | None): A color palette for Rich progress bars.
            console (Console | None): A Rich Console for progress display.
            use_cache (bool): Whether to use the local cache for hashing and
                metadata. Defaults to True.
            lazy (bool): If True, processes files via an iterator.
        """
        self.offline = offline or is_offline_mode()
        self._client = client
        self.warnings: list[str] = []
        self._file_class = file_class

        self.remote_collection_id: str | None = None
        self.remote_last_modified: datetime.datetime | None = None
        self.remote_file_count: int | None = None

        final_files: Sequence[LocalFile]
        final_source_info: dict | None = source_info

        if files is not None:
            final_files = files
        elif isinstance(source, str):
            path = source
            self.reader = MetadataReader(
                client=self._client, model_config=model_runner_pipeline, file_class=file_class, offline=self.offline
            )
            scan_files, self.warnings = self.reader.scan_directory(
                dir_path=source,
                recursive=recursive,
                return_errors=True,
                console=console,
                palette=palette,
                skip_cache=not use_cache,
                overwrite_cache=overwrite_cache,
                follow_symlinks=follow_symlinks,
                lazy=lazy,
            )
            final_files = scan_files
            if self.warnings:
                logger.warning(
                    f"Initialized collection from '{path}', but {len(self.warnings)} "
                    f"files could not be processed. Check the .warnings attribute for details."
                )
            if not final_source_info:
                final_source_info = {
                    "type": "local",
                    "path": path,
                    "recursive": recursive,
                    "scan_started_at": datetime.datetime.now(datetime.timezone.utc).isoformat(),
                }
        elif isinstance(source, list):
            final_files = source
        else:
            raise ValueError("Either 'source' or 'files' must be provided to LocalFileCollection.")

        super().__init__(files=final_files, source_info=final_source_info)
        self._is_populated = True

    def __iter__(self) -> Iterator[LocalFile]:
        return iter(self.files)

    def __getitem__(self, index: int) -> LocalFile:  # type: ignore[override]
        return self.files[index]

    def __add__(self, other: "_BaseFileCollection") -> "LocalFileCollection":
        """
        Combines two LocalFileCollection objects into a new one.
        """
        if not isinstance(other, LocalFileCollection):
            raise TypeError("Addition is only supported between two LocalFileCollection objects.")

        combined_files_map = {f.hash: f for f in self.files}
        combined_files_map.update({f.hash: f for f in other.files})

        new_source_info = {
            "type": "merged",
            "operation": "addition",
            "sources": [self.source_info, other.source_info],
        }
        return self.__class__(
            source=cast(list[LocalFile], list(combined_files_map.values())),
            source_info=new_source_info,
        )

    def __sub__(self, other: _BaseFileCollection) -> LocalFileCollection:
        """
        Creates a new LocalFileCollection by removing files present in the
        second collection from the first.
        """
        if not isinstance(cast(object, other), _BaseFileCollection):
            return NotImplemented

        other_hashes = {f.hash for f in other.files}
        resulting_files = [f for f in self.files if f.hash not in other_hashes]

        new_source_info = {
            "type": "merged",
            "operation": "subtraction",
            "sources": [self.source_info, other.source_info],
        }
        return self.__class__(source=cast(list[LocalFile], resulting_files), source_info=new_source_info)

    def add_tags(
        self,
        tags: list[dict | NewFileTag],
        api_key: str | None = None,
        console: Console | None = None,
        palette: dict | None = None,
    ) -> LocalFileCollection:
        """
        Adds one or more tags to every file in the collection.

        This method first validates all tags against the server in a single
        batch, then applies them locally to each file object. The local changes
        must be synchronized with DorsalHub by calling `.push()`.

        Args:
            tags (list[dict | NewFileTag]): A list of tags to add. Each tag can
                be a dictionary or a `NewFileTag` object.
            api_key (str | None): An optional API key for validation.
            console (Console | None): A Rich Console for progress display.
            palette (dict | None): A color palette for the progress bar.

        Returns:
            The `LocalFileCollection` instance for method chaining.
        """
        from dorsal.file.validators.file_record import NewFileTag

        if not self.files:
            logger.warning("Cannot add tags: the collection is empty.")
            return self

        if not tags:
            logger.warning("No tags provided to add.")
            return self

        if self._client is None:
            self._client = get_shared_dorsal_client(api_key=api_key)

        if self.offline:
            logger.info("Step 1/2: *SKIPPING* tag validation - Offline Mode")
        else:
            logger.info(f"Step 1/2: Validating {len(tags)} tags in a single batch...")
            try:
                tags_to_validate = [tag if isinstance(tag, NewFileTag) else NewFileTag(**tag) for tag in tags]
            except Exception as e:
                raise DorsalClientError(f"Failed to parse input tags: {e}") from e

            validation_result = self._client.validate_tag(file_tags=tags_to_validate, api_key=api_key)

            if not validation_result.valid:
                error_msg = validation_result.message or "Tag validation failed with no specific message."
                logger.error(f"Tag validation failed: {error_msg}")
                raise InvalidTagError(error_msg)

            logger.info("Tag validation successful.")

        logger.info(f"Step 2/2: Applying {len(tags_to_validate)} tags to {len(self.files)} files...")

        rich_progress = None
        iterator: Iterable[LocalFile]
        if is_jupyter_environment():
            iterator = tqdm(self.files, desc="Applying tags")
        elif console:
            from rich.progress import (
                Progress,
                BarColumn,
                TaskProgressColumn,
                MofNCompleteColumn,
                TextColumn,
                TimeElapsedColumn,
                TimeRemainingColumn,
            )
            from dorsal.cli.themes.palettes import DEFAULT_PALETTE

            active_palette = palette if palette is not None else DEFAULT_PALETTE
            progress_columns = (
                TextColumn(
                    "[progress.description]{task.description}",
                    style=active_palette.get("progress_description", "default"),
                ),
                BarColumn(bar_width=None, style=active_palette.get("progress_bar", "default")),
                TaskProgressColumn(style=active_palette.get("progress_percentage", "default")),
                MofNCompleteColumn(),
                TextColumn("•", style="dim"),
                TimeElapsedColumn(),
                TextColumn("•", style="dim"),
                TimeRemainingColumn(),
            )
            rich_progress = Progress(
                *progress_columns,
                console=console,
                redirect_stdout=True,
                transient=True,
                redirect_stderr=True,
            )
            task_id = rich_progress.add_task("Applying tags...", total=len(self.files))
            iterator = self.files
        else:
            iterator = self.files

        with rich_progress if rich_progress else open(os.devnull, "w"):
            for file in iterator:
                if isinstance(file, LocalFile):
                    for tag in tags_to_validate:
                        try:
                            file._add_local_tag(name=tag.name, value=tag.value)
                        except Exception as e:
                            logger.warning(f"Could not apply tag '{tag.name}' to file '{file.name}': {e}")
                if rich_progress:
                    rich_progress.update(task_id, advance=1)

        logger.info("Batch tagging complete.")
        return self

    def push(
        self,
        public: bool = False,
        api_key: str | None = None,
        console: "Console | None" = None,
        palette: dict | None = None,
        fail_fast: bool = True,
        strict: bool = False,
    ) -> dict:
        """Pushes all file records in the collection to DorsalHub for indexing.

        Args:
            public: If True, uploads as public records.
            api_key: Optional API key override.
            console: Rich console for progress display.
            palette: Color palette for progress display.
            fail_fast: If True (default), aborts immediately on the first batch error.
            strict: If True, raises PartialIndexingError the response contains any errors.

        Returns:
            dict: Summary of the push operation.

        Raises:
            ValueError: If public indexing is attempted with restricted media types.
            PartialIndexingError: If strict=True and partial failures are detected.
        """
        from dorsal.file.metadata_reader import MetadataReader
        from dorsal.file.validators.file_record import FileRecordStrict
        from dorsal.common.exceptions import PartialIndexingError

        if public:
            prohibited_files = []
            prohibited_tags = []
            for file in self.files:
                if not is_permitted_public_media_type(file.media_type):
                    name_repr = file.name or file.hash or "Unknown File"
                    prohibited_files.append(f"'{name_repr}' ({file.media_type})")

                if any(t.name == "label" for t in file.tags):
                    name_repr = file.name or file.hash or "Unknown File"
                    prohibited_tags.append(name_repr)

            if prohibited_files:
                limit = 5
                details = ", ".join(prohibited_files[:limit])
                if len(prohibited_files) > limit:
                    details += f" and {len(prohibited_files) - limit} others"

                raise ValueError(
                    f"The collection cannot be indexed publicly because it contains restricted media types: {details}."
                )

            if prohibited_tags:
                limit = 5
                details = ", ".join(prohibited_tags[:limit])
                if len(prohibited_tags) > limit:
                    details += f" and {len(prohibited_tags) - limit} others"

                raise ValueError(
                    f"The collection cannot be indexed publicly because "
                    f"it contains files with strictly private 'label' tags: {details}."
                )

        if self._client is None:
            self._client = get_shared_dorsal_client(api_key=api_key)

        reader = MetadataReader(client=self._client, offline=self.offline)

        records_to_upload = [f.model for f in self.files if isinstance(f.model, FileRecordStrict)]

        if not records_to_upload:
            logger.info("No valid records in the collection to push.")
            return {
                "total_records": 0,
                "processed": 0,
                "success": 0,
                "failed": 0,
                "batches": [],
                "errors": [],
            }

        self.push_results = reader.upload_records(
            records=records_to_upload,
            public=public,
            fail_fast=fail_fast,
            console=console,
            palette=palette,
        )

        if strict:
            failed_count = self.push_results.get("failed", 0)
            if failed_count > 0:
                errors = self.push_results.get("errors", [])
                if not errors:
                    errors = [f"{failed_count} records failed to index properly."]

                raise PartialIndexingError(
                    message=f"Collection push failed in strict mode. {failed_count} errors detected.",
                    summary=self.push_results,
                )

        return self.push_results

    def sync_with_remote(
        self,
        api_key: str | None = None,
        force: bool = False,
        poll_interval: int = 5,
        timeout: int | None = 300,
    ) -> dict:
        """
        Synchronizes the linked remote collection to exactly match this local collection.
        """
        if not self.remote_collection_id:
            raise DorsalError(
                "Synchronization requires a linked remote collection. Use `create_remote_collection()` to create and link one first."
            )

        if self._client is None:
            self._client = get_shared_dorsal_client(api_key=api_key)

        logger.debug(f"Starting synchronization for remote collection: {self.remote_collection_id}")

        logger.debug("Step 1/3: Performing pre-flight check...")
        remote_state = self._client.get_collection(self.remote_collection_id, api_key=api_key, hydrate=False)

        is_state_synced = (
            remote_state.collection.date_modified == self.remote_last_modified
            and remote_state.collection.file_count == self.remote_file_count
        )

        if not is_state_synced and not force:
            raise SyncConflictError(
                "Sync failed: The remote collection has been modified since the last synchronization. To proceed and overwrite the remote changes, run the command again with `force=True`."
            )
        elif not is_state_synced and force:
            logger.warning("`force=True` provided. Overwriting remote changes.")
        else:
            logger.debug("Pre-flight check passed. Remote collection is in expected state.")

        logger.info("Step 2/3: Pushing local file records to ensure they exist on the server...")
        is_remote_private = remote_state.collection.is_private
        public = not is_remote_private
        push_summary = self.push(public=public, api_key=api_key)

        num_to_push = cast(int, push_summary.get("total_records_to_push", 0))
        num_accepted = cast(int, push_summary.get("success", 0))
        if not is_remote_private and num_accepted != num_to_push:
            raise DorsalClientError(
                f"Sync aborted: Not all local files could be indexed publicly ({num_accepted}/{num_to_push}). Cannot sync with a public collection."
            )

        logger.info("Step 3/3: Sending complete hash list to the server for synchronization...")
        local_hashes = [f.hash for f in self.files if f.hash]

        sync_response = self._client.sync_collection_by_hash(
            collection_id=self.remote_collection_id,
            hashes=local_hashes,
            api_key=api_key,
            poll_interval=poll_interval,
            timeout=timeout,
        )

        final_remote_state = self._client.get_collection(self.remote_collection_id, api_key=api_key, hydrate=False)
        self.remote_last_modified = final_remote_state.collection.date_modified
        self.remote_file_count = final_remote_state.collection.file_count

        logger.info(
            f"Synchronization complete. Added: {sync_response.added_count}, Removed: {sync_response.removed_count}, Unchanged: {sync_response.unchanged_count}."
        )
        return sync_response.model_dump()

    def create_remote_collection(
        self,
        name: str,
        description: str | None = None,
        public: bool = False,
        api_key: str | None = None,
    ) -> "DorsalFileCollection":
        """
        Creates a new remote collection on DorsalHub, populates it with the
        files from this local collection, and links the two.
        """
        from dorsal.file.collection.remote import DorsalFileCollection

        if self._client is None:
            self._client = get_shared_dorsal_client(api_key=api_key)

        logger.info("Step 1/3: Pushing file records to DorsalHub...")
        push_summary = self.push(public=public, api_key=api_key)
        if cast(int, push_summary["success"]) == 0:
            raise DorsalClientError("No files were successfully indexed. Cannot create collection.")
        logger.info("File records pushed successfully.")

        logger.info(f"Step 2/3: Creating remote collection '{name}'...")

        source_paths = _get_source_paths(self.source_info)
        collection_source = {
            "caller": "dorsal.LocalFileCollection",
            "local_directories": source_paths,
            "comment": "Created via the Dorsal Python library.",
        }

        is_private = not public
        remote_collection_meta = self._client.create_collection(
            name=name,
            description=description,
            is_private=is_private,
            source=collection_source,
        )
        collection_id = remote_collection_meta.collection_id

        logger.info("Step 3/3: Adding files to the new collection...")
        file_hashes = [file.hash for file in self.files if file.hash][:API_MAX_BATCH_SIZE]
        if len(file_hashes) > API_MAX_BATCH_SIZE:
            logger.warning(
                "This collection exceeds the max batch size."
                f"Only the first {API_MAX_BATCH_SIZE} records will be included initially."
                "You should run `sync_with_remote` after the collection is created to add the rest."
            )
        add_response = self._client.add_files_to_collection(collection_id=collection_id, hashes=file_hashes)
        logger.info(
            f"Successfully added {add_response.added_count} files to collection '{name}'. "
            f"({add_response.duplicate_count} duplicates ignored)."
        )

        collection_response = self._client.get_collection(collection_id=collection_id, per_page=0, hydrate=False)
        remote_collection = collection_response.collection

        self.remote_collection_id = collection_id
        self.remote_last_modified = remote_collection.date_modified
        self.remote_file_count = remote_collection.file_count
        logger.info("Remote collection created and linked. Local state updated.")

        return DorsalFileCollection(collection_id=collection_id, client=self._client)

    def to_dict(
        self,
        by_alias: bool = True,
        exclude_none: bool = True,
        exclude: dict | set | None = None,
    ) -> dict:
        """
        Serializes the local collection to a dictionary, augmenting each file
        record with essential local filesystem attributes.
        """
        data = super().to_dict(by_alias=by_alias, exclude_none=exclude_none, exclude=exclude)

        for i, file_obj in enumerate(self.files):
            if i < len(data["results"]):
                if isinstance(file_obj, LocalFile):
                    data["results"][i]["local_attributes"] = {
                        "date_modified": file_obj.date_modified,
                        "date_created": file_obj.date_created,
                        "file_path": file_obj._file_path,
                    }
        return data



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import subprocess
import pytest
from unittest.mock import MagicMock, patch

from dorsal.registry import uninstaller
from dorsal.common.exceptions import DorsalError


@pytest.fixture
def mock_resolve_target():
    """Mock resolution to always return a safe package name."""
    with patch("dorsal.registry.uninstaller.resolve_target") as mock:
        mock.return_value = ("package", "dorsal-test-model")
        yield mock


@pytest.fixture
def mock_unregister_model():
    """Mock the config modification."""
    with patch("dorsal.registry.uninstaller.unregister_model") as mock:
        yield mock


@pytest.fixture
def mock_subprocess():
    """Mock the pip subprocess."""
    with patch("subprocess.Popen") as mock_popen:
        process_mock = MagicMock()
        process_mock.stdout = []
        process_mock.wait.return_value = 0
        mock_popen.return_value = process_mock
        yield mock_popen


def test_uninstall_full_success(mock_resolve_target, mock_unregister_model, mock_subprocess):
    """
    Scenario: The model exists in config AND is installed in pip.
    Result: Success.
    """

    result = uninstaller.uninstall_model_target("test-model")

    assert result == "dorsal-test-model"
    mock_unregister_model.assert_called_with(package_name="dorsal-test-model", scope="project")

    expected_cmd = [sys.executable, "-m", "pip", "uninstall", "-y", "dorsal-test-model"]
    mock_subprocess.assert_called_with(
        expected_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        encoding="utf-8",
        errors="replace",
        bufsize=1,
    )


def test_uninstall_config_only(mock_resolve_target, mock_unregister_model, mock_subprocess):
    """
    Scenario: The model is in config, but pip says "Skipping (not installed)".
    Result: Success (Config was cleaned up).
    """

    process_mock = mock_subprocess.return_value
    process_mock.stdout = ["Skipping dorsal-test-model as it is not installed\n"]

    result = uninstaller.uninstall_model_target("test-model")

    assert result == "dorsal-test-model"
    mock_unregister_model.assert_called_once()


def test_uninstall_pip_only(mock_resolve_target, mock_unregister_model, mock_subprocess):
    """
    Scenario: The model is NOT in config, but IS installed in pip.
    Result: Success (Package was removed).
    """

    mock_unregister_model.side_effect = KeyError("Model not found")

    result = uninstaller.uninstall_model_target("test-model")

    assert result == "dorsal-test-model"

    mock_unregister_model.assert_called_once()


def test_uninstall_noop_failure(mock_resolve_target, mock_unregister_model, mock_subprocess):
    """
    Scenario: Model NOT in config AND NOT installed in pip.
    Result: DorsalError (Nothing was done).
    """

    mock_unregister_model.side_effect = KeyError("Model not found")

    process_mock = mock_subprocess.return_value
    process_mock.stdout = ["Skipping dorsal-test-model as it is not installed\n"]

    with pytest.raises(DorsalError) as exc:
        uninstaller.uninstall_model_target("test-model")

    assert "Could not find model" in str(exc.value)
    assert "resolved as package: 'dorsal-test-model'" in str(exc.value)


def test_uninstall_pip_error_passthrough(mock_resolve_target, mock_unregister_model, mock_subprocess):
    """
    Scenario: Pip subprocess fails with a non-zero exit code.
    Result: Since config also failed, raise Error.
    """
    mock_unregister_model.side_effect = KeyError("Model not found")

    process_mock = mock_subprocess.return_value
    process_mock.wait.return_value = 1
    process_mock.stdout = ["Permission denied\n"]

    with pytest.raises(DorsalError) as exc:
        uninstaller.uninstall_model_target("test-model")

    assert "Could not find model" in str(exc.value)


def test_pip_streaming_logs_on_error(mock_subprocess, capsys):
    """
    Verify internal helper prints logs to stderr when pip fails.
    """

    uninstaller.logger.setLevel(50)

    process_mock = mock_subprocess.return_value
    process_mock.wait.return_value = 1
    process_mock.stdout = ["Fatal Error in Uninstall\n"]

    success = uninstaller._run_pip_uninstall_streaming("bad-pkg")

    assert success is False
    captured = capsys.readouterr()
    assert "Pip Uninstall Log" in captured.err
    assert "Fatal Error" in captured.err



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from typing import Annotated, Any, Self
import uuid

from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)

from dorsal.common.validators import String128, TString255
from dorsal.file.validators.common import QuickHash, SHA256Hash, TLSHash
from dorsal.file.utils.hashes import hash_string_validator, HashFunctionId

logger = logging.getLogger(__name__)

FILESIZE_UPPER_LIMIT = 2**50  # 1 Petabyte
RX_FILE_EXTENSION = re.compile(r"^\.[a-zA-Z0-9_!-]{1,16}$")

MediaTypeString = Annotated[str, Field(pattern=r"^\w+\/[-+.\w]+$", min_length=3, max_length=256)]
MediaTypePartString = Annotated[
    str, Field(pattern=r"^(?:\w+\/[-+.\w]+|\w+)$", min_length=3, max_length=256)
]  # `video/png` or just `video`


def validate_file_extension_or_null(value: Any) -> str | None:
    """Convert input to a lowercase string if it matches RX_FILE_EXTENSION.

    Return `None` for empty or unmatchings strings, or `None`

    """
    if value is None:
        return None

    try:
        str_value = str(value)

        if not str_value:
            return None

        if RX_FILE_EXTENSION.match(str_value):
            return str_value.lower()
        else:
            logger.debug(
                "File extension '%s' (original: '%s') does not match regex. Returning None.",
                str_value,
                value,
            )
            return None

    except (TypeError, AttributeError) as e:
        logger.debug(
            "Failed to process value for file extension validation (value: '%s', type: %s). Error: %s. Returning None.",
            value,
            type(value).__name__,
            e,
        )
        return None


FileExtension = Annotated[str | None, BeforeValidator(validate_file_extension_or_null)]


class FileCoreValidationModelHash(BaseModel):
    """
    Represents a single hash record with its type identifier and value.
    Validates the hash value against its specified type.
    """

    id: HashFunctionId
    value: String128

    @field_validator("value")
    @classmethod
    def validate_hash_value_format_and_normalize_case(cls, value: str, info: ValidationInfo) -> str:
        """
        Validates the hash string's format based on the 'id' (hash function).
        Normalizes the case of the hash string (e.g., TLSH to uppercase, others to lowercase).
        Raises ValueError if the hash function is unsupported or the hash value is invalid.
        """
        hash_function_id = info.data.get("id")
        if not hash_function_id:
            raise ValueError("Hash function 'id' is missing, cannot validate hash value.")  # pragma: no cover

        if not hash_string_validator.is_supported_hash_function(hash_function_id):
            raise ValueError(f"Hash function '{hash_function_id}' is not currently supported.")

        normalized_hash = hash_string_validator.get_valid_hash(candidate_string=value, hash_function=hash_function_id)
        if normalized_hash is None:
            raise ValueError(f"Value '{value}' is not a valid hash for function '{hash_function_id}'.")
        return normalized_hash


class FileCoreValidationModel(BaseModel):
    """
    Base Pydantic model for validating the core metadata extracted by `FileCoreAnnotationModel`.
    This record forms the foundational data about a file.
    """

    hash: SHA256Hash | None = None
    quick_hash: QuickHash | None = None
    similarity_hash: TLSHash | None = None
    name: TString255
    extension: FileExtension | None = None
    size: int = Field(ge=0, lt=FILESIZE_UPPER_LIMIT)
    media_type: MediaTypeString
    all_hashes: list[FileCoreValidationModelHash] | None = Field(default=None, min_length=2, max_length=4)
    all_hash_ids: dict[HashFunctionId, str] | None = Field(default=None)

    @computed_field  # type: ignore
    @property
    def media_type_prefix(self) -> str:
        return self.media_type.split("/")[0]

    @model_validator(mode="after")
    def populate_and_validate_all_hash_ids(self) -> Self:
        """
        Populates `all_hash_ids`, overwrites top-level hashes, and performs validation checks:
        - Populates `all_hash_ids` from `all_hashes`.
        - Overwrites `quick_hash` and `similarity_hash` from `all_hashes`.
        - Ensures SHA256 and BLAKE3 hashes are present.
        - Verifies no duplicate hash IDs.
        - Confirms the primary `self.hash` matches the SHA256 value in `all_hashes`.
        """
        if self.all_hashes is not None:
            temp_all_hash_ids: dict[HashFunctionId, str] = {}
            for hash_record in self.all_hashes:
                if hash_record.id in temp_all_hash_ids:
                    raise ValueError(f"Duplicate hash ID '{hash_record.id}' found in record.all_hashes.")
                temp_all_hash_ids[hash_record.id] = hash_record.value
            self.all_hash_ids = temp_all_hash_ids

            if "QUICK" in self.all_hash_ids:
                self.quick_hash = self.all_hash_ids["QUICK"]

            if "TLSH" in self.all_hash_ids:
                self.similarity_hash = self.all_hash_ids["TLSH"]

            if len(set(self.all_hash_ids.values())) < len(self.all_hash_ids):
                logger.warning(
                    "Potential issue: Identical hash values found for different hash IDs in all_hashes. Record hash: %s",
                    self.hash,
                )

            if "SHA-256" not in self.all_hash_ids:
                raise ValueError("SHA-256 file hash missing from record.all_hashes.")
            if "BLAKE3" not in self.all_hash_ids:
                raise ValueError("BLAKE3 file hash missing from record.all_hashes.")
            if self.hash != self.all_hash_ids["SHA-256"]:
                raise ValueError("Record 'hash' (primary SHA256) does not match SHA-256 value in record.all_hashes.")
        return self


class FileCoreValidationModelStrict(FileCoreValidationModel):
    """
    A validated `FileCoreValidationModel` - i.e. contains 'BLAKE3' validation hash within `all_hashes`.
    """

    hash: SHA256Hash
    all_hashes: list[FileCoreValidationModelHash] = Field(min_length=2, max_length=4)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from unittest.mock import Mock, MagicMock, patch
import logging
import sys
import zipfile
import xml.etree.ElementTree as ET


from dorsal.file.preprocessing.office import extract_docx_text

# --- Helpers ---

DOC_NS = 'xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"'


def build_docx_xml(body_content: str) -> bytes:
    """
    Wraps content in a valid OpenXML document structure with namespaces.
    Returns bytes, as zipfile.read() returns bytes.
    """
    xml = (
        f'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        f"<w:document {DOC_NS}>"
        f"<w:body>{body_content}</w:body>"
        f"</w:document>"
    )
    return xml.encode("utf-8")


@pytest.fixture
def mock_zip():
    """
    Patches zipfile.ZipFile to avoid file I/O.
    Returns the mock instance that the context manager yields.
    """
    with patch("zipfile.ZipFile") as MockZipFile:
        mock_instance = MockZipFile.return_value
        # Setup context manager protocol
        mock_instance.__enter__.return_value = mock_instance
        mock_instance.__exit__.return_value = None
        yield mock_instance


# --- Tests ---


class TestExtractDocxText:
    def test_single_page_simple_text(self, mock_zip):
        """Test basic extraction of paragraphs and text."""
        # <w:p><w:r><w:t>Hello World</w:t></w:r></w:p>
        content = "<w:p><w:r><w:t>Hello World</w:t></w:r></w:p>"

        mock_zip.read.return_value = build_docx_xml(content)

        pages = extract_docx_text("dummy.docx")

        # The function logic:
        # <w:p> adds "\n"
        # <w:t> adds "Hello World"
        # Result "Hello World" (stripped)
        assert len(pages) == 1
        assert pages[0] == "Hello World"
        mock_zip.read.assert_called_with("word/document.xml")

    def test_explicit_page_break(self, mock_zip):
        """Test splitting text on hard page breaks."""
        # Page 1: "Page One"
        # Break: <w:br w:type="page"/>
        # Page 2: "Page Two"
        content = '<w:p><w:t>Page One</w:t></w:p><w:p><w:br w:type="page"/></w:p><w:p><w:t>Page Two</w:t></w:p>'
        mock_zip.read.return_value = build_docx_xml(content)

        pages = extract_docx_text("dummy.docx")

        assert len(pages) == 2
        assert pages[0] == "Page One"
        assert pages[1] == "Page Two"

    def test_formatting_elements(self, mock_zip):
        """Test handling of tabs and line breaks."""
        # "Line\nBreak" and "Tab\tCharacter"
        # <w:br/> is a line break, <w:tab/> is a tab
        content = (
            "<w:p><w:t>Line</w:t><w:br/><w:t>Break</w:t></w:p><w:p><w:t>Tab</w:t><w:tab/><w:t>Character</w:t></w:p>"
        )
        mock_zip.read.return_value = build_docx_xml(content)

        pages = extract_docx_text("dummy.docx")

        assert len(pages) == 1
        text = pages[0]
        assert "Line\nBreak" in text
        assert "Tab\tCharacter" in text

    def test_handles_invalid_zip(self, mock_zip):
        """Test that BadZipFile is caught and logs error."""
        # Make the constructor raise, simulating a corrupt file
        # We need to patch the constructor directly, but since our fixture
        # already mocked the class, we can set the side_effect on the mock class calls
        # However, it's cleaner to patch specific behavior for this test.

        # Resetting logic: The function does `with zipfile.ZipFile(...)`.
        # If construction fails:
        with patch("zipfile.ZipFile", side_effect=zipfile.BadZipFile("Bad zip")):
            pages = extract_docx_text("corrupt.docx")

        assert pages == []

    def test_handles_missing_document_xml(self, mock_zip):
        """Test that KeyError (missing file in zip) is caught."""
        # Simulation: zip file opens fine, but 'word/document.xml' is missing
        mock_zip.read.side_effect = KeyError("word/document.xml not found")

        pages = extract_docx_text("empty_zip.docx")

        assert pages == []

    def test_handles_malformed_xml(self, mock_zip):
        """Test that XML parse errors are caught."""
        # Return invalid XML bytes
        mock_zip.read.return_value = b"<w:document> <UnclosedTag> </w:document>"

        pages = extract_docx_text("bad_xml.docx")

        assert pages == []

    def test_empty_text_nodes(self, mock_zip):
        """Test that <w:t> tags with no content (None) don't crash."""
        # <w:t/> with no text
        content = "<w:p><w:t/></w:p>"
        mock_zip.read.return_value = build_docx_xml(content)

        pages = extract_docx_text("dummy.docx")

        # Should return empty string page (or nothing depending on stripping)
        # ".strip()" on empty list join result "" is ""
        # If the page has content "\n" (from <w:p>), it strips to ""
        assert len(pages) == 0 or (len(pages) == 1 and pages[0] == "")



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import logging
from typing import Any, TYPE_CHECKING

from rich.console import Group, RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.syntax import Syntax
from rich.bar import Bar
from rich.rule import Rule
from rich.json import JSON

if TYPE_CHECKING:
    from dorsal.file.configs.model_runner import RunModelResult
    from dorsal.client.validators import FileAnnotationResponse, FileAnnotationGroupResponse

logger = logging.getLogger(__name__)


def create_model_result_panel(
    result: "RunModelResult | FileAnnotationResponse | FileAnnotationGroupResponse",
    title: str,
    file_name: str,
    palette: dict[str, str],
    max_length: int = 10,
) -> Panel:
    """
    Dispatcher that renders a rich Panel for a model run result.
    """
    from dorsal.file.configs.model_runner import RunModelResult
    from dorsal.client.validators import FileAnnotationResponse, FileAnnotationGroupResponse

    logger.debug(f"Rendering model result panel for '{file_name}' (Target: {title})")
    logger.debug(f"Result object type: {type(result).__name__}")

    data: dict[str, Any] | None = None
    group_info = ""
    schema_id = ""

    if isinstance(result, RunModelResult):
        logger.debug("Extracting record directly from RunModelResult")
        data = result.record
        schema_id = getattr(result, "schema_id", "") or ""

    elif isinstance(result, FileAnnotationGroupResponse):
        logger.debug("Extracting record from FileAnnotationGroupResponse")
        if result.group.annotations and result.group.annotations[0].record:
            data = result.group.annotations[0].record.model_dump()
            schema_id = getattr(result.group.annotations[0], "schema_id", "")
        group_info = f" (Group of {len(result.group.annotations)})"

    elif isinstance(result, FileAnnotationResponse):
        logger.debug("Extracting record from FileAnnotationResponse")
        data = result.record
        schema_id = getattr(result, "schema_id", "")

    logger.debug(f"Extracted schema_id: '{schema_id}'")

    if not data:
        logger.debug("No record data found in result. Returning Empty Result panel.")
        return Panel(
            Text("No record data returned.", style=palette.get("error", "red")),
            title=f"[{palette.get('panel_title_error', 'red')}] Empty Result{group_info}[/]",
            border_style=palette.get("panel_border_error", "red"),
        )

    content: RenderableType
    schema_type: str

    if schema_id in RENDERER_REGISTRY:
        logger.debug(f"Found specific renderer for schema_id '{schema_id}' in RENDERER_REGISTRY.")
        render_func, schema_type = RENDERER_REGISTRY[schema_id]
        content = render_func(data, palette, max_length)
    else:
        logger.debug(f"No specific renderer found for schema_id '{schema_id}'. Falling back to Raw JSON Output.")
        content = JSON.from_data(data)
        schema_type = "Raw Output"

    title_style = palette.get("panel_title_success", "green")
    border_style = palette.get("panel_border_success", "green")

    header = Table.grid(expand=True)
    header.add_column(justify="left")
    header.add_column(justify="right")
    header.add_row(
        Text(title, style=palette.get("section_title", "bold")), Text(f"{file_name}", style=palette.get("info", "dim"))
    )

    logger.debug("Successfully constructed Panel layout.")
    return Panel(
        Group(header, Rule(style=palette.get("info", "dim")), Text(""), content),
        title=f"[{title_style}]{schema_type} Result{group_info}[/]",
        border_style=border_style,
        expand=False,
    )


def _extract_color(style_str: str, default: str) -> str:
    """Helper to safely extract just a color name from a rich style string."""
    if not style_str:
        return default
    words = style_str.replace("on ", "").split()
    for w in reversed(words):
        if w not in ("bold", "dim", "italic", "underline", "strike", "blink", "reverse"):
            return w
    return default


def _score_bar(score: float, palette: dict[str, str], width: int = 20) -> Bar:
    """Helper to create a visual score bar respecting the theme palette."""
    if score > 0.7:
        style_str = palette.get("success", "green")
    elif score > 0.4:
        style_str = palette.get("warning", "yellow")
    else:
        style_str = palette.get("error", "red")

    color = _extract_color(style_str, "default")
    bgcolor = _extract_color(palette.get("table_row_alt", "bright_black"), "black")

    return Bar(size=width, begin=0, end=score, color=color, bgcolor=bgcolor)


def _render_arxiv(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_arxiv")
    renderables: list[RenderableType] = []

    title = data.get("title", "Untitled")
    title = " ".join(title.split())
    arxiv_id = data.get("arxiv_id", "Unknown ID")
    version = data.get("version", "")

    if version:
        version_str = version if version.startswith("v") else f"v{version}"
        id_str = f"{arxiv_id} ({version_str})"
    else:
        id_str = arxiv_id

    renderables.append(Text(title, style=palette.get("section_title", "bold")))
    renderables.append(Text(id_str, style=palette.get("info", "dim")))

    authors = data.get("authors", [])
    if authors:
        # We can apply max_length to the authors list too!
        display_authors = authors[:max_length]
        author_str = ", ".join(display_authors)
        if len(authors) > max_length:
            author_str += f", ... (+{len(authors) - max_length})"
        renderables.append(Text(author_str, style=f"{palette.get('primary_value_alt', 'cyan')} italic"))

    renderables.append(Text(""))

    abstract = data.get("abstract", "")
    if abstract:
        renderables.append(
            Panel(
                abstract,
                title=f"[{palette.get('panel_title_info', 'dim')}]Abstract[/]",
                border_style=palette.get("panel_border_info", "dim"),
            )
        )
        renderables.append(Text(""))

    grid = Table.grid(padding=(0, 2))
    grid.add_column(style=palette.get("key", "dim"))
    grid.add_column(style=palette.get("primary_value", "default"))

    if url := data.get("url"):
        grid.add_row("URL", Text(url, style=palette.get("link", "underline")))
    if categories := data.get("categories"):
        grid.add_row("Categories", ", ".join(categories))
    if doi := data.get("doi"):
        grid.add_row("DOI", doi)
    if journal := data.get("journal_ref"):
        grid.add_row("Journal", journal)
    if license_str := data.get("license"):
        grid.add_row("License", license_str)

    if grid.row_count > 0:
        renderables.append(grid)

    return Group(*renderables)


def _render_classification(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_classification")
    renderables: list[RenderableType] = []

    grid = Table.grid(padding=(0, 2))
    grid.add_column(style=palette.get("key", "dim"))
    grid.add_column(style=palette.get("primary_value", "bold"))
    if target := data.get("target"):
        grid.add_row("Target:", str(target))
    if producer := data.get("producer"):
        grid.add_row("Producer:", str(producer))
    if grid.row_count > 0:
        renderables.extend([grid, Text("")])

    labels = data.get("labels", [])
    if labels:
        table = Table(box=None, padding=(0, 2), show_header=True)
        table.add_column("Label", style=palette.get("table_header", "bold"))
        table.add_column("Score", justify="right", style=palette.get("primary_value", "cyan"))
        table.add_column("Confidence", width=20)

        labels_sorted = sorted(labels, key=lambda x: x.get("score", 0), reverse=True)
        for item in labels_sorted[:max_length]:
            score = item.get("score", 0)
            table.add_row(item.get("label", "Unknown"), f"{score:.4f}", _score_bar(score, palette))

        if len(labels_sorted) > max_length:
            table.add_row(f"... and {len(labels_sorted) - max_length} more", "", "")
        renderables.append(table)
    else:
        vocab = data.get("vocabulary", [])
        if vocab:
            renderables.append(Text(f"Vocabulary: {', '.join(vocab[:max_length])}", style=palette.get("info", "dim")))
        elif v_url := data.get("vocabulary_url"):
            renderables.append(Text(f"Vocabulary URL: {v_url}", style=palette.get("link", "underline")))

    if desc := data.get("score_explanation"):
        renderables.extend([Text(""), Text(f"Score: {desc}", style=f"{palette.get('info', 'dim')} italic")])

    return Group(*renderables)


def _render_entity_extraction(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_entity_extraction")
    renderables: list[RenderableType] = []

    info_grid = Table.grid(padding=(0, 2))
    info_grid.add_column(style=palette.get("key", "dim"))
    info_grid.add_column(style=palette.get("primary_value", "bold"))
    if val := data.get("producer"):
        info_grid.add_row("Producer:", str(val))
    if val := data.get("unit"):
        info_grid.add_row("Unit:", str(val))
    if info_grid.row_count > 0:
        renderables.extend([info_grid, Text("")])

    entities = data.get("entities", [])
    if not entities:
        renderables.append(Text("No entities found.", style=palette.get("info", "dim")))
        return Group(*renderables)

    table = Table(box=None, padding=(0, 2), show_header=True)
    table.add_column("Label", style=palette.get("table_header", "bold"))
    table.add_column("Concept", style=palette.get("info", "dim"))
    table.add_column("Text", style=palette.get("primary_value", "cyan"))
    table.add_column("Normalized Value", style=palette.get("primary_value_alt", "yellow"))
    table.add_column("Score", justify="right", style=palette.get("primary_value", "cyan"))

    for ent in entities[:max_length]:
        val = str(ent.get("value")) if ent.get("value") is not None else "-"
        table.add_row(
            ent.get("label", "UNK"),
            ent.get("concept", "-"),
            ent.get("text", "")[:50],
            val[:30],
            f"{ent.get('score', 0):.2f}" if "score" in ent else "-",
        )

    renderables.extend([table, Text(f"Extracted {len(entities)} entities.", style=palette.get("info", "dim"))])
    return Group(*renderables)


def _render_object_detection(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_object_detection")
    renderables: list[RenderableType] = []

    info_grid = Table.grid(padding=(0, 2))
    info_grid.add_column(style=palette.get("key", "dim"))
    info_grid.add_column(style=palette.get("primary_value", "bold"))
    if val := data.get("producer"):
        info_grid.add_row("Producer:", str(val))
    if val := data.get("unit"):
        info_grid.add_row("Unit:", str(val))
    if info_grid.row_count > 0:
        renderables.extend([info_grid, Text("")])

    objects = data.get("objects", [])
    table = Table(box=None, padding=(0, 2), show_header=True)
    table.add_column("Label", style=palette.get("table_header", "bold"))
    table.add_column("Score", justify="right", style=palette.get("primary_value", "cyan"))
    table.add_column("Location", style=palette.get("primary_value", "cyan"))

    for obj in objects[:max_length]:
        loc = "Unknown"
        if box := obj.get("box"):
            loc = f"Box [x:{float(box.get('x', 0)):.1f}, y:{float(box.get('y', 0)):.1f}]"
        elif poly := obj.get("polygon"):
            loc = f"Poly ({len(poly)} pts)"

        table.add_row(obj.get("label", "Unknown"), f"{obj.get('score', 0):.2f}" if "score" in obj else "-", loc)

    renderables.extend([table, Text(f"Found {len(objects)} objects.", style=palette.get("info", "dim"))])
    return Group(*renderables)


def _render_embedding(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_embedding")
    renderables: list[RenderableType] = []

    grid = Table.grid(padding=(0, 2))
    grid.add_column(style=palette.get("key", "dim"))
    grid.add_column(style=palette.get("primary_value", "bold"))

    if model := data.get("model"):
        grid.add_row("Model", str(model))
    if target := data.get("target"):
        grid.add_row("Target", str(target))

    vector = data.get("vector")
    if isinstance(vector, list):
        dim = len(vector)
        grid.add_row("Type", "Dense Array")
        grid.add_row("Dimensions", str(dim))
        vec_preview = "[" + ", ".join(f"{v:.4f}" for v in vector[:max_length])
        if dim > max_length:
            vec_preview += f", ... (+{dim - max_length} more)"
        vec_preview += "]"
    elif isinstance(vector, dict):
        dim = vector.get("dimensions", "Unknown")
        indices = vector.get("indices", [])
        grid.add_row("Type", "Sparse Object")
        grid.add_row("Dimensions", str(dim))
        grid.add_row("Non-zero Elements", str(len(indices)))
        vec_preview = f"Indices: {indices[:max_length]}...\nValues: {vector.get('values', [])[:max_length]}..."
    else:
        vec_preview = "Unknown vector format."

    renderables.extend(
        [
            grid,
            Text(""),
            Panel(
                vec_preview,
                title=f"[{palette.get('panel_title_info', 'dim')}]Vector Data[/]",
                border_style=palette.get("panel_border_info", "dim"),
            ),
        ]
    )

    return Group(*renderables)


def _render_audio_transcription(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_audio_transcription")
    renderables: list[RenderableType] = []

    info_grid = Table.grid(padding=(0, 2))
    info_grid.add_column(style=palette.get("key", "dim"))
    info_grid.add_column(style=palette.get("primary_value", "bold"))

    for key, label in [
        ("track_id", "Track ID"),
        ("producer", "Producer"),
        ("language", "Language"),
        ("duration", "Duration (s)"),
    ]:
        if val := data.get(key):
            info_grid.add_row(f"{label}:", str(val))

    if info_grid.row_count > 0:
        renderables.extend([info_grid, Text("")])

    if full_text := data.get("text"):
        renderables.extend(
            [
                Panel(
                    full_text,
                    title=f"[{palette.get('panel_title_info', 'dim')}]Full Transcription[/]",
                    border_style=palette.get("panel_border_info", "dim"),
                ),
                Text(""),
            ]
        )

    segments = data.get("segments", [])
    if segments:
        seg_table = Table(box=None, padding=(0, 1), show_header=True)
        seg_table.add_column("Time", style=palette.get("info", "dim"))
        seg_table.add_column("Speaker", style=palette.get("table_header", "bold"))
        seg_table.add_column("Text", style=palette.get("primary_value", "cyan"))

        for seg in segments[:max_length]:
            start = seg.get("start_time", 0)
            end = seg.get("end_time", 0)
            speaker_obj = seg.get("speaker", {})
            speaker = speaker_obj.get("name") or str(speaker_obj.get("id", "Unknown"))

            seg_table.add_row(f"{start:.1f}-{end:.1f}s", speaker, seg.get("text", ""))

        renderables.append(seg_table)
        if len(segments) > max_length:
            renderables.append(
                Text(f"... {len(segments) - max_length} more segments", style=palette.get("info", "dim"))
            )

    return Group(*renderables)


def _render_document_extraction(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_document_extraction")
    blocks = data.get("blocks", [])
    stats: dict[str, int] = {}
    for b in blocks:
        btype = b.get("block_type", "unknown")
        stats[btype] = stats.get(btype, 0) + 1

    summary = ", ".join(f"{k}: {v}" for k, v in stats.items())
    renderables: list[RenderableType] = []

    info_grid = Table.grid(padding=(0, 2))
    info_grid.add_column(style=palette.get("key", "dim"))
    info_grid.add_column(style=palette.get("primary_value", "bold"))
    for key, label in [("extraction_type", "Type"), ("producer", "Producer"), ("unit", "Unit")]:
        if val := data.get(key):
            info_grid.add_row(f"{label}:", str(val))

    if info_grid.row_count > 0:
        renderables.append(info_grid)

    renderables.extend([Text(f"Summary: {summary}", style=palette.get("info", "dim")), Text("")])

    text_blocks = [b for b in blocks if b.get("text")]
    for b in text_blocks[:max_length]:
        renderables.append(
            Panel(
                b.get("text", ""),
                title=f"[{palette.get('panel_title_info', 'dim')}]Block {b.get('id', '')[:8]} (Page {b.get('page_number', '?')})[/]",
                border_style=palette.get("panel_border_info", "dim"),
            )
        )

    if len(text_blocks) > max_length:
        renderables.append(Text(f"... {len(text_blocks) - max_length} more blocks", style=palette.get("info", "dim")))

    return Group(*renderables)


def _render_llm_output(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_llm_output")
    renderables: list[RenderableType] = []

    renderables.append(Text(f"Model: {data.get('model', 'Unknown Model')}", style=palette.get("section_title", "bold")))
    if lang := data.get("language"):
        renderables.append(Text(f"Language: {lang}", style=palette.get("info", "dim")))

    prompt = data.get("prompt", "")
    if len(prompt) > 200:
        prompt = prompt[:200] + "..."
    renderables.append(
        Panel(
            prompt,
            title=f"[{palette.get('panel_title_info', 'dim')}]Input Prompt[/]",
            border_style=palette.get("panel_border_info", "dim"),
            height=5,
        )
    )

    response = data.get("response_data", "")
    response_render: RenderableType
    if isinstance(response, str) and (response.strip().startswith("{") or response.strip().startswith("[")):
        response_render = Syntax(response, "json", word_wrap=True)
    else:
        response_render = Text(str(response))

    renderables.append(
        Panel(
            response_render,
            title=f"[{palette.get('panel_title_success', 'bold green')}]Response[/]",
            border_style=palette.get("panel_border_success", "green"),
        )
    )

    meta_grid = Table.grid(padding=(0, 2))
    meta_grid.add_column(style=palette.get("key", "dim"))
    meta_grid.add_column(style=palette.get("info", "dim"))

    if usage := data.get("generation_metadata", {}).get("usage"):
        stats = f"Tokens: {usage.get('total_tokens')} (Prompt: {usage.get('prompt_tokens')}, Compl: {usage.get('completion_tokens')})"
        meta_grid.add_row("Usage:", stats)

    if params := data.get("generation_params", {}):
        param_str = ", ".join(f"{k}={v}" for k, v in params.items() if k in ("temperature", "top_p", "max_tokens"))
        if param_str:
            meta_grid.add_row("Params:", param_str)

    if meta_grid.row_count > 0:
        renderables.append(meta_grid)

    return Group(*renderables)


def _render_regression(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_regression")
    points = data.get("points", [])
    target = data.get("target", "Unknown Target")
    unit = data.get("unit", "")

    renderables: list[RenderableType] = [
        Text(f"Regression Target: {target}", style=palette.get("section_title", "bold"))
    ]
    if producer := data.get("producer"):
        renderables.append(Text(f"Producer: {producer}", style=palette.get("info", "dim")))
    renderables.append(Text(""))

    table = Table(box=None, padding=(0, 2), show_header=True)
    table.add_column("Timestamp", style=palette.get("info", "dim"))
    table.add_column(f"Value ({unit})" if unit else "Value", style=palette.get("table_header", "bold"))
    table.add_column("Statistic", style=palette.get("primary_value", "cyan"))
    table.add_column("Interval", style=palette.get("primary_value", "cyan"))

    for p in points[:max_length]:
        val = p.get("value")
        val_str = f"{val:.4f}" if isinstance(val, float) else str(val)

        stat = p.get("statistic", "point")
        if stat == "quantile" and "quantile_level" in p:
            stat = f"quantile ({p['quantile_level']})"

        interval = ""
        if p.get("interval_lower") is not None and p.get("interval_upper") is not None:
            interval = f"[{p['interval_lower']:.2f}, {p['interval_upper']:.2f}]"

        table.add_row(p.get("timestamp", "-"), val_str, stat, interval)

    renderables.append(table)

    if len(points) > max_length:
        renderables.append(Text(f"... {len(points) - max_length} more points", style=palette.get("info", "dim")))

    return Group(*renderables)


def _render_geolocation(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_geolocation")
    geo = data.get("geometry") or {}
    props = data.get("properties") or {}

    grid = Table.grid(padding=(0, 2))
    grid.add_column(style=palette.get("key", "dim"))
    grid.add_column(style=palette.get("primary_value", "cyan"))

    grid.add_row("Type", geo.get("type", "Null Geometry"))
    if coords := geo.get("coordinates"):
        coord_str = str(coords)
        grid.add_row("Coordinates", coord_str[:100] + "..." if len(coord_str) > 100 else coord_str)

    if props:
        renderables: list[RenderableType] = [
            grid,
            Rule(style=palette.get("info", "dim")),
            Text("Properties", style=palette.get("section_title", "bold")),
        ]

        # Apply max_length to properties dictionary
        prop_items = list(props.items())
        for k, v in prop_items[:max_length]:
            if v is not None:
                renderables.append(Text(f"{k}: {v}", style=palette.get("primary_value", "cyan")))

        if len(prop_items) > max_length:
            renderables.append(
                Text(f"... {len(prop_items) - max_length} more properties", style=palette.get("info", "dim"))
            )

        return Group(*renderables)

    return grid


def _render_generic(data: dict[str, Any], palette: dict[str, str], max_length: int) -> RenderableType:
    logger.debug("Executing _render_generic")
    kv_data = data.get("data", {})
    renderables: list[RenderableType] = [
        Text(data.get("description", "Generic Data"), style=f"{palette.get('info', 'dim')} italic")
    ]
    if producer := data.get("producer"):
        renderables.append(Text(f"Producer: {producer}", style=palette.get("info", "dim")))
    renderables.append(Text(""))

    table = Table(box=None, padding=(0, 2))
    table.add_column("Key", style=palette.get("key", "dim"))
    table.add_column("Value", style=palette.get("primary_value", "cyan"))

    kv_items = list(kv_data.items())
    for k, v in kv_items[:max_length]:
        table.add_row(k, str(v))

    renderables.append(table)

    if len(kv_items) > max_length:
        renderables.append(Text(f"... {len(kv_items) - max_length} more properties", style=palette.get("info", "dim")))

    return Group(*renderables)


RENDERER_REGISTRY = {
    "dorsal/arxiv": (_render_arxiv, "ArXiv Record"),
    "open/audio-transcription": (_render_audio_transcription, "Audio Transcription"),
    "open/classification": (_render_classification, "Classification"),
    "open/document-extraction": (_render_document_extraction, "Document Extraction"),
    "open/embedding": (_render_embedding, "Embedding"),
    "open/entity-extraction": (_render_entity_extraction, "Entity Extraction"),
    "open/geolocation": (_render_geolocation, "Geolocation"),
    "open/llm-output": (_render_llm_output, "LLM Output"),
    "open/object-detection": (_render_object_detection, "Object Detection"),
    "open/regression": (_render_regression, "Regression"),
    "open/generic": (_render_generic, "Generic Data"),
}



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer
import pathlib
import json
import logging
from typing import Annotated

from rich.panel import Panel
from rich.text import Text
from rich.markup import escape

logger = logging.getLogger(__name__)


def identify_file_cmd(
    ctx: typer.Context,
    path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="The path to the local file to identify.",
        ),
    ],
    secure: Annotated[
        bool,
        typer.Option(
            "--secure",
            help="Use SHA-256 to hash the file, instead of the default QUICK hash.",
        ),
    ] = False,
    use_cache: Annotated[
        bool,
        typer.Option(
            "--use-cache",
            help="Force the use of the cache, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    skip_cache: Annotated[
        bool,
        typer.Option(
            "--skip-cache",
            help="Bypass the local cache and re-process the file, overriding any global setting.",
            rich_help_panel="Cache Options",
        ),
    ] = False,
    json_output: Annotated[
        bool,
        typer.Option(
            "--json",
            help="Output the full metadata as a raw JSON object to stdout for scripting.",
        ),
    ] = False,
):
    """
    Identifies a local file by its hash, by checking DorsalHub.
    """
    from dorsal.api.file import identify_file
    from dorsal.cli.views.file import create_file_info_panel
    from dorsal.common.cli import (
        get_rich_console,
        exit_cli,
        EXIT_CODE_ERROR,
        determine_use_cache_value,
    )
    from dorsal.common.exceptions import AuthError, DorsalClientError, DorsalOfflineError, NotFoundError

    console = get_rich_console()
    palette = ctx.obj["palette"]

    if use_cache and skip_cache:
        exit_cli(
            code=EXIT_CODE_ERROR,
            message="Error: --use-cache and --skip-cache flags cannot be used together.",
        )

    use_cache_value = determine_use_cache_value(use_cache=use_cache, skip_cache=skip_cache)

    if not json_output:
        console.print(f"🔎 Identifying file [{palette['primary_value']}]{escape(path.name)}[/]...")

    try:
        is_quick_mode = not secure

        file_record_dict = identify_file(
            file_path=str(path),
            quick=is_quick_mode,
            use_cache=use_cache_value,
            mode="dict",
        )

        if json_output:
            console.print(json.dumps(file_record_dict, indent=2, default=str, ensure_ascii=False))
            raise typer.Exit(0)

        panel = create_file_info_panel(
            record_dict=file_record_dict,
            title="✅ File Identified",
            private=None,
            palette=palette,
        )
        console.print(panel)

    except typer.Exit:
        raise
    except NotFoundError as e:
        if json_output:
            error_payload = {
                "success": False,
                "error": "Not Found",
                "detail": e.message,
            }
            console.print(json.dumps(error_payload, indent=2))
        else:
            message = Text.assemble(
                ("This file has not been indexed to DorsalHub.\n\n", "default"),
                ("Why not (privately) index it yourself, by running:\n", "default"),
                (
                    f'dorsal file push "{escape(str(path))}"\n\n',
                    f"bold {palette.get('primary_value', 'default')}",
                ),
                ("You can also index it publicly, by running:\n", "default"),
                (
                    f'dorsal file push "{escape(str(path))}" --public',
                    f"bold {palette.get('primary_value', 'default')}",
                ),
            )
            console.print(
                Panel(
                    message,
                    title=f"[{palette.get('panel_title_warning', 'bold yellow')}]⚠️ Not Found[/]",
                    border_style=palette.get("panel_border_warning", "yellow"),
                    expand=False,
                )
            )
        exit_cli(code=EXIT_CODE_ERROR)
    except DorsalOfflineError:
        raise
    except AuthError:
        raise

    except DorsalClientError as err:
        if json_output:
            error_payload = {
                "success": False,
                "error": "API Error",
                "detail": err.message,
            }
            console.print(json.dumps(error_payload, indent=2))
        else:
            exit_cli(code=EXIT_CODE_ERROR, message=f"API Error: {err.message}")
        exit_cli(code=EXIT_CODE_ERROR)
    except Exception as err:
        logger.exception(f"CLI 'identify' command failed for path {path}.")
        if json_output:
            error_payload = {
                "success": False,
                "error": "Unexpected Error",
                "detail": str(err),
            }
            console.print(json.dumps(error_payload, indent=2))
        else:
            exit_cli(code=EXIT_CODE_ERROR, message=f"An unexpected error occurred: {err}")
        exit_cli(code=EXIT_CODE_ERROR)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
from typer.testing import CliRunner
from unittest.mock import MagicMock
from rich.panel import Panel


from dorsal.cli import app
from dorsal.common.exceptions import DorsalClientError, AuthError

runner = CliRunner()

TEST_FILE_PATH = "tests/data/valid.txt"


@pytest.fixture
def mock_push_cmd(mocker):
    """
    Mocks dependencies for the `file push` command.
    """
    mocker.patch("dorsal.common.cli.determine_use_cache_value", return_value=True)

    mock_local_file_class = mocker.patch("dorsal.file.dorsal_file.LocalFile")

    mock_instance = mock_local_file_class.return_value

    mock_api_response = MagicMock()
    mock_api_response.success = 1
    mock_api_response.error = 0
    mock_api_response.results = [MagicMock()]
    mock_api_response.results[0].hash = "mock_pushed_hash"
    mock_api_response.model_dump.return_value = {
        "success": 1,
        "error": 0,
        "results": [{"hash": "mock_pushed_hash"}],
    }

    mock_instance.push.return_value = mock_api_response

    return {
        "local_file_class": mock_local_file_class,
        "local_file_instance": mock_instance,
    }


def test_push_success_private_panel(mock_rich_console, mock_push_cmd):
    """Tests a successful private push (the default)."""
    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH])

    assert result.exit_code == 0
    mock_push_cmd["local_file_instance"].push.assert_called_once_with(public=False, strict=False)

    panel_output = mock_rich_console.print.call_args_list[1].args[0]
    assert isinstance(panel_output, Panel)
    assert "Push Complete" in str(panel_output.title)
    assert "mock_pushed_hash" in str(panel_output.renderable)


def test_push_success_public_panel(mock_rich_console, mock_push_cmd):
    """Tests a successful public push using the --public flag."""
    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH, "--public"])

    assert result.exit_code == 0
    mock_push_cmd["local_file_instance"].push.assert_called_once_with(public=True, strict=False)

    panel_output = mock_rich_console.print.call_args_list[1].args[0]
    assert isinstance(panel_output, Panel)
    assert "Push Complete" in str(panel_output.title)


def test_push_success_json_output(mock_rich_console, mock_push_cmd):
    """Tests a successful push with --json output."""
    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH, "--json"])

    assert result.exit_code == 0
    json_output_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_output_str)
    assert data["success"] == 1
    assert data["results"][0]["hash"] == "mock_pushed_hash"


def test_push_api_failure_panel(mock_rich_console, mock_push_cmd):
    """Tests a failed push due to an API-side reason."""
    mock_api_response = mock_push_cmd["local_file_instance"].push.return_value
    mock_api_response.success = 0
    mock_api_response.error = 1
    mock_api_response.results[0].annotations[0].detail = "File already indexed"

    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH])

    assert result.exit_code == 0
    panel_output = mock_rich_console.print.call_args_list[1].args[0]
    assert isinstance(panel_output, Panel)
    assert "Push Failed" in str(panel_output.title)
    assert "File already indexed" in str(panel_output.renderable)


def test_push_auth_error(mock_push_cmd):
    """
    Tests that the command bubbles up AuthError (does not swallow it).
    """
    mock_push_cmd["local_file_instance"].push.side_effect = AuthError("test_msg")

    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH])

    assert result.exit_code != 0
    assert isinstance(result.exception, AuthError)


def test_push_client_error(mock_push_cmd):
    """Tests that a DorsalClientError is handled correctly."""
    mock_push_cmd["local_file_instance"].push.side_effect = DorsalClientError("Connection timeout")
    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH])

    assert result.exit_code != 0
    assert "API Error: Connection timeout" in result.output


def test_push_cache_flag_conflict():
    """Tests that using both --use-cache and --skip-cache fails."""
    result = runner.invoke(app, ["file", "push", TEST_FILE_PATH, "--use-cache", "--skip-cache"])
    assert result.exit_code != 0
    assert "Error: --use-cache and --skip-cache flags cannot be used together" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from types import SimpleNamespace
from pydantic import BaseModel, ValidationError as PydanticValidationError
from dorsal.file.file_annotator import FileAnnotator
from dorsal.common.exceptions import (
    AnnotationConfigurationError,
    AnnotationExecutionError,
    AnnotationImportError,
    AnnotationValidationError,
    ModelRunnerError,
)
from dorsal.file.configs.model_runner import ModelRunnerPipelineStep
from dorsal.file.validators.file_record import GenericFileAnnotation, Annotation
from dorsal.common.model import AnnotationModel, AnnotationModelSource, AnnotationManualSource
from dorsal.common.validators import CallableImportPath


class MockAnnotationModel(AnnotationModel):
    id = "mock_model"

    def main(self):
        pass


class MockPydanticModel(BaseModel):
    field: str


class BadModelNoId:
    pass


@pytest.fixture
def annotator():
    return FileAnnotator()


@pytest.fixture
def mock_runner(mocker):
    runner = mocker.MagicMock()
    # Valid hash is 64 chars (sha256)
    valid_hash = "a" * 64

    # Return a "dumb" object that mimics RunModelResult structure
    # to bypass strict Pydantic validation of Source types during testing.
    # We use a Model source here as that is what the Runner typically returns.
    source_obj = AnnotationModelSource(id="mock_model", version="1.0")

    result = SimpleNamespace(
        name="mock_model",
        schema_id="test/schema",
        record={"file_hash": valid_hash, "data": {"foo": "bar"}},
        source=source_obj,
        error=None,
    )
    runner.run_single_model.return_value = result
    return runner


# --- Tests: _execute ---


def test_execute_success(annotator, mock_runner):
    res = annotator._execute(
        model_runner=mock_runner,
        annotation_model=MockAnnotationModel,
        validation_model=None,
        file_path="test.file",
        schema_id="test/schema",
        options={"opt": 1},
    )
    assert res.record["file_hash"] == "a" * 64
    mock_runner.run_single_model.assert_called_once()


def test_execute_runner_raises_exception(annotator, mock_runner):
    mock_runner.run_single_model.side_effect = ModelRunnerError("Simulated Crash")
    with pytest.raises(AnnotationExecutionError) as exc:
        annotator._execute(mock_runner, MockAnnotationModel, None, "path", "id", {})
    assert "Execution failed for model" in str(exc.value)


def test_execute_runner_returns_error_string(annotator, mock_runner):
    mock_runner.run_single_model.return_value = SimpleNamespace(
        name="mock_model",
        schema_id="test/schema",
        record=None,
        source=AnnotationManualSource(id="error"),
        error="Logic Error inside Model",
    )
    with pytest.raises(AnnotationExecutionError) as exc:
        annotator._execute(mock_runner, MockAnnotationModel, None, "path", "id", {})
    assert "returned an error: Logic Error inside Model" in str(exc.value)


def test_execute_runner_returns_nothing(annotator, mock_runner):
    mock_runner.run_single_model.return_value = SimpleNamespace(
        name="mock_model",
        schema_id="test/schema",
        record=None,
        source=AnnotationManualSource(id="empty"),
        error=None,
    )
    with pytest.raises(AnnotationExecutionError) as exc:
        annotator._execute(mock_runner, MockAnnotationModel, None, "path", "id", {})
    assert "returned no record and no error" in str(exc.value)


# --- Tests: annotate_file_using_model_and_validator ---


def test_annotate_direct_success(annotator, mock_runner, mocker):
    mocker.patch("dorsal.file.file_annotator.is_valid_dataset_id_or_schema_id", return_value=True)

    result = annotator.annotate_file_using_model_and_validator(
        file_path="test.file",
        model_runner=mock_runner,
        annotation_model_cls=MockAnnotationModel,
        schema_id="test/direct",
        private=False,
    )
    assert isinstance(result, Annotation)
    assert result.record.file_hash == "a" * 64


def test_annotate_direct_missing_schema(annotator, mock_runner):
    with pytest.raises(AnnotationConfigurationError):
        annotator.annotate_file_using_model_and_validator(
            file_path="f",
            model_runner=mock_runner,
            annotation_model_cls=MockAnnotationModel,
            schema_id=None,
            private=False,  # type: ignore
        )


def test_annotate_direct_bad_model_class(annotator, mock_runner):
    with pytest.raises(AnnotationConfigurationError) as exc:
        annotator.annotate_file_using_model_and_validator(
            file_path="f",
            model_runner=mock_runner,
            annotation_model_cls=BadModelNoId,  # type: ignore
            schema_id="test/schema",
            private=False,
        )
    assert "missing a required, non-empty 'id'" in str(exc.value)


# --- Tests: annotate_file_using_pipeline_step ---


def test_annotate_pipeline_step_dict_success(annotator, mock_runner, mocker):
    mocker.patch("dorsal.file.file_annotator.is_valid_dataset_id_or_schema_id", return_value=True)

    # Updated mock to intercept the new resolution function
    mocker.patch("dorsal.file.file_annotator.resolve_pipeline_step_models", return_value=(MockAnnotationModel, None))

    step_config = {
        "annotation_model": {"module": "some.module", "name": "ModelClass"},
        "schema_id": "test/pipeline",
        "options": {"foo": "bar"},
    }

    result = annotator.annotate_file_using_pipeline_step(
        file_path="f", model_runner=mock_runner, pipeline_step=step_config, private=True
    )

    assert result.record.file_hash == "a" * 64
    assert result.private is True


def test_annotate_pipeline_import_error(annotator, mock_runner, mocker):
    # Updated mock to throw the expected AnnotationImportError
    mocker.patch(
        "dorsal.file.file_annotator.resolve_pipeline_step_models",
        side_effect=AnnotationImportError("Failed to import model/validator"),
    )

    cip = CallableImportPath(module="foo", name="Bar")
    step = ModelRunnerPipelineStep(annotation_model=cip, schema_id="test/schema")

    with pytest.raises(AnnotationImportError) as exc:
        annotator.annotate_file_using_pipeline_step(
            file_path="f", model_runner=mock_runner, pipeline_step=step, private=False
        )
    assert "Failed to import model/validator" in str(exc.value)


def test_annotate_pipeline_type_error(annotator, mock_runner, mocker):
    # Updated mock to throw the expected AnnotationImportError
    mocker.patch(
        "dorsal.file.file_annotator.resolve_pipeline_step_models",
        side_effect=AnnotationImportError("Failed to import model/validator"),
    )

    cip = CallableImportPath(module="foo", name="Bad")
    step = ModelRunnerPipelineStep(annotation_model=cip, schema_id="test/schema")

    with pytest.raises(AnnotationImportError) as exc:
        annotator.annotate_file_using_pipeline_step(
            file_path="f", model_runner=mock_runner, pipeline_step=step, private=False
        )

    assert "Failed to import model/validator" in str(exc.value)


# --- Tests: validate_manual_annotation ---


def test_validate_manual_pydantic_success(annotator):
    data = {"field": "valid_value"}
    res = annotator.validate_manual_annotation(data, validator=MockPydanticModel)
    assert res["field"] == "valid_value"


def test_validate_manual_pydantic_failure(annotator):
    data = {"field": 123}
    with pytest.raises(AnnotationValidationError):
        annotator.validate_manual_annotation(data, validator=MockPydanticModel)


def test_validate_manual_no_validator(annotator):
    data = {"any": "thing"}
    res = annotator.validate_manual_annotation(data, validator=None)
    assert res == data


def test_validate_manual_unsupported_validator(annotator):
    with pytest.raises(AnnotationConfigurationError):
        annotator.validate_manual_annotation({}, validator=str)  # type: ignore


# --- Tests: make_manual_annotation ---


def test_make_manual_annotation_success(annotator, mocker):
    mocker.patch("dorsal.file.file_annotator.is_valid_dataset_id_or_schema_id", return_value=True)
    mocker.patch("dorsal.file.file_annotator.apply_linter")

    valid_hash = "a" * 64
    data = {"file_hash": valid_hash, "data": {}}

    res = annotator.make_manual_annotation(
        annotation=data, schema_id="manual/test", source_id="user input", private=False
    )

    assert res.record.file_hash == valid_hash
    assert res.source.id == "user input"


def test_make_manual_annotation_force(annotator, mocker):
    spy_val = mocker.spy(annotator, "validate_manual_annotation")

    valid_hash = "a" * 64
    valid_data = {"file_hash": valid_hash, "data": {}}

    annotator.make_manual_annotation(
        annotation=valid_data, schema_id="forced/schema", source_id=None, private=False, force=True
    )

    spy_val.assert_not_called()


def test_make_annotation_invalid_schema_id(annotator, mocker):
    mocker.patch("dorsal.file.file_annotator.is_valid_dataset_id_or_schema_id", return_value=False)

    with pytest.raises(AnnotationConfigurationError):
        annotator._make_annotation(
            validated_annotation={}, schema_id="invalid/id", schema_version=None, source={}, private=False
        )


def test_make_annotation_wrapper_lookup(annotator, mocker):
    mocker.patch("dorsal.file.file_annotator.is_valid_dataset_id_or_schema_id", return_value=True)

    class SpecialAnnotation(Annotation):
        pass

    mocker.patch.dict("dorsal.file.file_annotator.CORE_MODEL_ANNOTATION_WRAPPERS", {"special/id": SpecialAnnotation})

    valid_hash = "b" * 64
    valid_data = {"file_hash": valid_hash, "data": {}}

    # Use 'Manual' source to satisfy union in _make_annotation
    source_dict = {"type": "Manual", "id": "test"}

    res = annotator._make_annotation(
        validated_annotation=valid_data, schema_id="special/id", schema_version=None, source=source_dict, private=False
    )

    assert isinstance(res, SpecialAnnotation)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
from typer.testing import CliRunner
from unittest.mock import MagicMock


from dorsal.cli import app

runner = CliRunner()


@pytest.fixture
def mock_prune_cache_cmd(mocker):
    """Mocks dependencies for the `cache prune` command."""
    mock_get_cache = mocker.patch("dorsal.session.get_shared_cache")

    mock_cache_instance = MagicMock()
    mock_get_cache.return_value = mock_cache_instance

    return {"get_cache": mock_get_cache, "cache_instance": mock_cache_instance}


def test_prune_cache_with_stale_records(mock_rich_console, mock_prune_cache_cmd):
    """Tests the prune command when stale records are found and removed."""
    mock_prune_cache_cmd["cache_instance"].prune.return_value = (10, 100)

    result = runner.invoke(app, ["cache", "prune"])

    assert result.exit_code == 0
    mock_prune_cache_cmd["cache_instance"].prune.assert_called_once()

    success_msg = mock_rich_console.print.call_args.args[0]
    assert "Prune complete. Removed 10 of 100 records" in success_msg


def test_prune_cache_no_stale_records(mock_rich_console, mock_prune_cache_cmd):
    """Tests the prune command when no stale records are found."""
    mock_prune_cache_cmd["cache_instance"].prune.return_value = (0, 50)

    result = runner.invoke(app, ["cache", "prune"])

    assert result.exit_code == 0
    mock_prune_cache_cmd["cache_instance"].prune.assert_called_once()

    info_msg = mock_rich_console.print.call_args.args[0]
    assert "Prune complete. No stale records found out of 50 scanned" in info_msg


def test_prune_cache_json_output(mock_rich_console, mock_prune_cache_cmd):
    """Tests the --json output flag."""
    mock_prune_cache_cmd["cache_instance"].prune.return_value = (10, 100)
    result = runner.invoke(app, ["cache", "prune", "--json"])

    assert result.exit_code == 0
    mock_rich_console.print.assert_called_once()

    json_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_str)

    assert data == {
        "total_records_scanned": 100,
        "stale_records_removed": 10,
    }


def test_prune_cache_exception_handling(mock_prune_cache_cmd):
    """Tests that a generic exception is handled gracefully."""
    mock_prune_cache_cmd["cache_instance"].prune.side_effect = Exception("Database is corrupt")
    result = runner.invoke(app, ["cache", "prune"])
    assert result.exit_code != 0
    assert "An error occurred while pruning the cache: Database is corrupt" in result.output



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import json
from pathlib import Path
from unittest.mock import MagicMock, patch
from typer.testing import CliRunner
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from dorsal.cli import app
from dorsal.common.auth import APIKeySource, APIKeyDetails

runner = CliRunner()


@pytest.fixture
def mock_config_app(mocker):
    """
    Mocks all backend dependencies for the `dorsal config` commands.
    """
    mocker.patch("dorsal.api.config.load_config", return_value=({}, Path("/fake/path/dorsal.toml")))
    mocker.patch("dorsal.api.config.get_global_config_path", return_value=Path("/fake/global/dorsal.toml"))

    mocker.patch("dorsal.api.config.get_email_from_config", return_value="test@example.com")
    mocker.patch("dorsal.api.config.get_theme_from_config", return_value="default")

    mocker.patch(
        "dorsal.api.config.get_api_key_details",
        return_value={
            "source": APIKeySource.PROJECT,
            "value": "fake_key",
            "path": Path("/fake/path/dorsal.toml"),
        },
    )

    mocker.patch("dorsal.api.config.constants.BASE_URL", "https://api.dorsalhub.test/v1")

    mocker.patch("dorsal.cli.themes.palettes.BUILT_IN_PALETTES", {"default": {}, "dark": {}})
    mocker.patch("dorsal.cli.themes.palettes._load_custom_palettes", return_value={})
    mocker.patch("dorsal.common.auth.write_theme_to_config")

    mocker.patch("dorsal.common.constants.LOCAL_DORSAL_DIR", Path("/fake/dorsal/dir"))


@patch("rich.table.Table.grid")
def test_show_config_logged_in(mock_table_grid, mock_rich_console, mock_config_app):
    """Tests `config show` output when the user is logged in."""
    mock_table = MagicMock()
    mock_table_grid.return_value = mock_table

    result = runner.invoke(app, ["config", "show"])

    assert result.exit_code == 0
    all_rows_text = ""
    for call in mock_table.add_row.call_args_list:
        all_rows_text += " ".join(str(arg) for arg in call.args)

    assert "test@example.com" in all_rows_text
    assert "project config" in all_rows_text


@patch("rich.table.Table.grid")
def test_show_config_logged_out(mock_table_grid, mocker, mock_rich_console, mock_config_app):
    """Tests `config show` output when the user is logged out."""
    mock_table = MagicMock()
    mock_table_grid.return_value = mock_table

    mocker.patch(
        "dorsal.api.config.get_api_key_details",
        return_value={"source": APIKeySource.NONE, "value": None, "path": None},
    )
    mocker.patch("dorsal.api.config.get_email_from_config", return_value=None)

    result = runner.invoke(app, ["config", "show"])

    assert result.exit_code == 0
    all_rows_text = ""
    for call in mock_table.add_row.call_args_list:
        all_rows_text += " ".join(str(arg) for arg in call.args)

    assert "Not Set" in all_rows_text
    assert "N/A (run 'dorsal auth login')" in all_rows_text


def test_show_config_json_output(mock_rich_console, mock_config_app):
    """Tests `config show --json` output."""
    result = runner.invoke(app, ["config", "show", "--json"])

    assert result.exit_code == 0
    json_output_str = mock_rich_console.print.call_args.args[0]
    data = json.loads(json_output_str)

    assert data["api_key_set"] is True
    assert "project config" in data["api_key_source"]
    assert data["logged_in_user"] == "test@example.com"


def test_list_themes_built_in_only(mock_rich_console, mock_config_app):
    """Tests `config theme list` with only built-in themes."""
    result = runner.invoke(app, ["config", "theme", "list"])

    assert result.exit_code == 0
    panel_output = mock_rich_console.print.call_args.args[0]
    group_renderables = panel_output.renderable.renderables

    all_text_content = []
    for r in group_renderables:
        if isinstance(r, Panel):
            all_text_content.append(str(r.title))
        else:
            all_text_content.append(str(r))

    combined_text = "".join(all_text_content)

    assert "Built-in Themes" in combined_text
    assert "Custom Themes" not in combined_text


def test_list_themes_with_custom(mocker, mock_rich_console, mock_config_app):
    """Tests `config theme list` with custom themes present."""
    mocker.patch(
        "dorsal.cli.themes.palettes._load_custom_palettes",
        return_value={"my_theme": {}},
    )

    result = runner.invoke(app, ["config", "theme", "list"])

    assert result.exit_code == 0
    panel_output = mock_rich_console.print.call_args.args[0]
    group_renderables = panel_output.renderable.renderables

    all_text_content = []
    for r in group_renderables:
        if isinstance(r, Panel):
            all_text_content.append(str(r.title))
        elif isinstance(r, str):
            all_text_content.append(r)
        elif isinstance(r, Text):
            all_text_content.append(r.plain)

    combined_text = "".join(all_text_content)

    assert "Built-in Themes" in combined_text
    assert "Custom Themes" in combined_text
    assert "my_theme" in combined_text


def test_set_theme_success(mock_rich_console, mock_config_app):
    """Tests successfully setting a valid theme."""
    result = runner.invoke(app, ["config", "theme", "set", "dark"])

    assert result.exit_code == 0
    from dorsal.common.auth import write_theme_to_config

    write_theme_to_config.assert_called_once_with("dark")
    success_message = str(mock_rich_console.print.call_args.args[0])
    assert "Default theme set to 'dark'" in success_message


def test_set_theme_not_found(mock_rich_console, mock_config_app):
    """Tests setting a theme that does not exist."""
    result = runner.invoke(app, ["config", "theme", "set", "invalid_theme"])

    assert result.exit_code != 0
    error_message = str(mock_rich_console.print.call_args.args[0])
    assert "Theme 'invalid_theme' not found" in error_message


def test_set_theme_os_error(mocker, mock_rich_console, mock_config_app):
    """Tests handling of an OSError when writing the theme config."""
    mocker.patch(
        "dorsal.common.auth.write_theme_to_config",
        side_effect=OSError("Permission denied"),
    )

    result = runner.invoke(app, ["config", "theme", "set", "dark"])

    assert result.exit_code != 0
    assert "Error saving configuration: Permission denied" in result.output


@pytest.fixture
def mock_pipeline_api(mocker):
    """Mocks the API functions used by the pipeline CLI."""
    return {
        "show": mocker.patch("dorsal.api.config.show_model_pipeline"),
        "remove_idx": mocker.patch("dorsal.api.config.remove_model_by_index"),
        "remove_name": mocker.patch("dorsal.api.config.remove_model_by_name"),
        "activate_idx": mocker.patch("dorsal.api.config.activate_model_by_index"),
        "activate_name": mocker.patch("dorsal.api.config.activate_model_by_name"),
        "deactivate_idx": mocker.patch("dorsal.api.config.deactivate_model_by_index"),
        "deactivate_name": mocker.patch("dorsal.api.config.deactivate_model_by_name"),
        "get_steps": mocker.patch("dorsal.api.config.get_model_pipeline"),
        "import_callable": mocker.patch("dorsal.common.validators.import_callable"),
    }


def test_pipeline_show_empty(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline show` when empty."""
    mock_pipeline_api["show"].return_value = []

    result = runner.invoke(app, ["config", "pipeline", "show"])

    assert result.exit_code == 0
    assert "pipeline is currently empty" in str(mock_rich_console.print.call_args.args[0].renderable)


def test_pipeline_show_populated(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline show` with data."""
    mock_pipeline_api["show"].return_value = [
        {
            "index": 0,
            "status": "Base (Locked)",
            "name": "BaseModel",
            "module": "dorsal.annotation_models",
            "schema_id": "base",
            "dependencies": "None",
        },
        {
            "index": 1,
            "status": "Deactivated",
            "name": "CustomModel",
            "module": "my.pkg",
            "schema_id": "custom",
            "dependencies": "pdf",
        },
    ]

    result = runner.invoke(app, ["config", "pipeline", "show"])

    assert result.exit_code == 0
    assert "Annotation Model Pipeline" in str(mock_rich_console.print.call_args.args[0].title)


@pytest.mark.parametrize(
    "command, api_idx, api_name, verb",
    [
        ("remove", "remove_idx", "remove_name", "removed"),
        ("activate", "activate_idx", "activate_name", "activated"),
        ("deactivate", "deactivate_idx", "deactivate_name", "deactivated"),
    ],
)
def test_pipeline_actions(mock_rich_console, mock_pipeline_api, command, api_idx, api_name, verb):
    """Generic test for remove/activate/deactivate by index and name."""

    result_idx = runner.invoke(app, ["config", "pipeline", command, "1"])
    assert result_idx.exit_code == 0
    mock_pipeline_api[api_idx].assert_called_with(index=1)
    assert f"Successfully {verb} model at index 1" in str(mock_rich_console.print.call_args.args[0])

    result_name = runner.invoke(app, ["config", "pipeline", command, "MyModel"])
    assert result_name.exit_code == 0
    mock_pipeline_api[api_name].assert_called_with(name="MyModel")
    assert f"Successfully {verb} model 'MyModel'" in str(mock_rich_console.print.call_args.args[0])


def test_pipeline_action_error(mock_rich_console, mock_pipeline_api):
    """Test error handling in pipeline actions."""
    mock_pipeline_api["remove_idx"].side_effect = IndexError("Index out of range")

    result = runner.invoke(app, ["config", "pipeline", "remove", "99"])

    assert result.exit_code != 0
    assert "Error: Index out of range" in str(mock_rich_console.print.call_args.args[0])


def test_pipeline_check_clean(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline check` when everything is fine."""

    step1 = MagicMock()
    step1.annotation_model = "valid.path"
    step1.validation_model = None

    mock_pipeline_api["get_steps"].return_value = [MagicMock(), step1]

    result = runner.invoke(app, ["config", "pipeline", "check"])

    assert result.exit_code == 0
    assert "All pipeline models are importable" in str(mock_rich_console.print.call_args.args[0])


def test_pipeline_check_broken_with_fix(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline check --fix` with broken models."""

    step_broken = MagicMock()
    step_broken.annotation_model.name = "BrokenModel"

    mock_pipeline_api["get_steps"].return_value = [MagicMock(), step_broken]

    mock_pipeline_api["import_callable"].side_effect = ImportError("Missing module")

    result = runner.invoke(app, ["config", "pipeline", "check", "--fix"])

    assert result.exit_code == 0

    output_text = "".join(str(c.args[0]) for c in mock_rich_console.print.call_args_list)
    assert "Found 1 broken models" in output_text
    assert "Removed broken model 'BrokenModel'" in output_text

    mock_pipeline_api["remove_idx"].assert_called_with(index=1)


def test_pipeline_check_broken_no_fix(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline check` (no fix) exits with error."""
    step_broken = MagicMock()
    step_broken.annotation_model.name = "BrokenModel"
    mock_pipeline_api["get_steps"].return_value = [MagicMock(), step_broken]
    mock_pipeline_api["import_callable"].side_effect = ImportError("Missing module")

    result = runner.invoke(app, ["config", "pipeline", "check"])

    assert result.exit_code != 0

    output_text = "".join(str(c.args[0]) for c in mock_rich_console.print.call_args_list)
    assert "--fix" in output_text
    assert "automatically remove" in output_text

    mock_pipeline_api["remove_idx"].assert_not_called()


def test_pipeline_show_json_output(mock_rich_console, mock_pipeline_api):
    """Tests `config pipeline show --json` output."""
    mock_pipeline_api["show"].return_value = [
        {
            "index": 0,
            "status": "Base (Locked)",
            "name": "BaseModel",
            "module": "dorsal.annotation_models",
            "schema_id": "base",
            "dependencies": "None",
        },
        {
            "index": 1,
            "status": "Active",
            "name": "MyCustomModel",
            "module": "custom.pkg",
            "schema_id": "custom_schema",
            "dependencies": ["base"],
        },
    ]

    result = runner.invoke(app, ["config", "pipeline", "show", "--json"])

    assert result.exit_code == 0

    json_output_str = mock_rich_console.print.call_args.args[0]

    data = json.loads(json_output_str)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]["name"] == "BaseModel"
    assert data[1]["schema_id"] == "custom_schema"



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typer

from dorsal.cli.dir_app.push_dir_cmd import push_directory
from dorsal.cli.dir_app.scan_dir_cmd import scan_directory
from dorsal.cli.dir_app.info_dir_cmd import info_directory
from dorsal.cli.dir_app.duplicates_dir_cmd import duplicates_dir

app = typer.Typer(
    name="dir",
    help="Scan, manage and index local files and directories.",
    no_args_is_help=True,
)

app.command(name="push")(push_directory)
app.command(name="scan")(scan_directory)
app.command(name="info")(info_directory)
app.command(name="duplicates")(duplicates_dir)



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import pytest
from rich.console import Console
from rich.table import Table
from dorsal.cli.views.file import create_file_info_panel, _build_annotation_table

DUMMY_FILE_RECORD = {
    "hash": "a" * 64,
    "validation_hash": "b" * 64,
    "quick_hash": "c" * 64,
    "similarity_hash": "d" * 64,
    "tags": [
        {"name": "project", "value": "apollo", "private": False, "id": "123"},
        {"name": "status", "value": "secret", "private": True},
    ],
    "annotations": {
        "file/base": {"record": {"name": "test_file.txt", "size": 1024, "media_type": "text/plain"}},
        "file/pdf": {"record": {"page_count": 5, "author": "Test Author"}},
        "something/anything": [
            {
                "source": {"type": "manual", "id": "user_1"},
                "date_modified": "2023-01-01T12:00:00Z",
                "url": "file/hash/123",
                "id": "anno_1",
            }
        ],
    },
    "local_filesystem": {"full_path": "/tmp/test_file.txt", "date_modified": datetime.datetime.now().isoformat()},
}

DUMMY_PALETTE = {
    "panel_border": "blue",
    "panel_border_alt": "red",
    "panel_title": "bold blue",
    "panel_title_alt": "bold red",
    "primary_value": "white",
    "primary_value_alt": "yellow",
    "key": "cyan",
    "hash_value": "green",
    "section_title": "bold white",
    "access_private": "red",
    "access_public": "green",
    "tag_private": "red",
    "tag_public": "blue",
    "tag_subtext": "dim",
    "info": "dim",
}


def test_create_file_info_panel_public(capsys):
    panel = create_file_info_panel(
        record_dict=DUMMY_FILE_RECORD, title="Test File", private=False, palette=DUMMY_PALETTE
    )

    console = Console()
    with console.capture() as capture:
        console.print(panel)

    output = capture.get()
    assert "Test File" in output
    assert "test_file.txt" in output
    assert "apollo" in output
    assert "page_count" in output
    assert "manual (user_1)" in output


def test_create_file_info_panel_private():
    """Smoke test for private file panel generation."""
    panel = create_file_info_panel(
        record_dict=DUMMY_FILE_RECORD, title="Private File", private=True, palette=DUMMY_PALETTE
    )
    console = Console()
    with console.capture() as capture:
        console.print(panel)

    output = capture.get()
    assert "Private Record" in output

    assert "Private File" in output


def test_create_file_info_panel_from_cache():
    """Test cache source labeling."""
    panel = create_file_info_panel(
        record_dict=DUMMY_FILE_RECORD, title="Cached File", private=False, palette=DUMMY_PALETTE, source="cache"
    )
    console = Console()
    with console.capture() as capture:
        console.print(panel)

    output = capture.get()
    assert "from cache" in output


def test_create_file_info_panel_minimal_data():
    """Test rendering with minimal data (empty dictionaries) to check for crashes."""
    minimal_record = {"hash": "123"}
    panel = create_file_info_panel(
        record_dict=minimal_record,
        title="Empty File",
        private=None,
        palette=DUMMY_PALETTE,
    )
    console = Console()
    with console.capture() as capture:
        console.print(panel)
    output = capture.get()
    assert "SHA-256" in output
    assert "No tags found" in output


def _render_table_to_string(table: Table) -> str:
    """Helper to render a Rich Table to a string for assertion."""
    console = Console()
    with console.capture() as capture:
        console.print(table)
    return capture.get()


def test_build_annotation_table_stub_full():
    """Test stub rendering with all expected fields."""
    table = Table()
    table.add_column("Key")
    table.add_column("Value")

    stub_data = {
        "source": {"type": "manual", "id": "user_42"},
        "date_modified": "2024-05-20T14:30:00Z",
        "id": "anno_123",
        "url": "api/v1/somehash/annotations",
    }

    _build_annotation_table("test_key", table, stub_data, is_stub=True)
    output = _render_table_to_string(table)

    assert "manual (user_42)" in output
    assert "2024-05-20 14:30" in output
    assert "anno_123" in output
    assert "dorsal annotation get somehash anno_123" in output


def test_build_annotation_table_stub_minimal():
    """Test stub rendering with missing optional fields."""
    table = Table()
    table.add_column("Key")
    table.add_column("Value")

    stub_data = {}

    _build_annotation_table("test_key", table, stub_data, is_stub=True)
    output = _render_table_to_string(table)

    assert "none (none)" in output
    assert "Modified" not in output
    assert "To View" not in output


def test_build_annotation_table_dict_logic():
    """Test nested dicts, lists, skipped keys, and empty values."""
    table = Table()
    table.add_column("Key")
    table.add_column("Value")

    data = {
        "file_hash": "this_should_be_skipped",
        "empty_string": "",
        "none_value": None,
        "empty_list": [],
        "simple_key": "simple_value",
        "nested_dict": {"inner_key": "inner_value"},
        "simple_list": [1, 2, "three"],
        "complex_list": [{"list_dict_key": "val1"}, {"list_dict_key2": "val2"}],
    }

    _build_annotation_table("test_key", table, data, is_stub=False)
    output = _render_table_to_string(table)

    assert "this_should_be_skipped" not in output
    assert "empty_string" not in output
    assert "none_value" not in output
    assert "empty_list" not in output

    assert "simple_key:" in output
    assert "simple_value" in output

    assert "nested_dict:" in output
    assert "inner_key:" in output
    assert "inner_value" in output

    assert "simple_list:" in output
    assert "1, 2, three" in output

    assert "complex_list:" in output
    assert "list_dict_key:" in output
    assert "val1" in output
    assert "list_dict_key2:" in output
    assert "val2" in output


def test_build_annotation_table_display_fields():
    """Test that display_fields correctly filters out unwanted keys recursively."""
    table = Table()
    table.add_column("Key")
    table.add_column("Value")

    data = {"keep_me": "yes", "drop_me": "no", "nested": {"keep_inner": "yes", "drop_inner": "no"}}

    _build_annotation_table("test_key", table, data, is_stub=False, display_fields={"keep_me", "nested", "keep_inner"})
    output = _render_table_to_string(table)

    assert "keep_me:" in output
    assert "drop_me:" not in output
    assert "nested:" in output
    assert "keep_inner:" in output
    assert "drop_inner:" not in output


def test_build_annotation_table_list_root():
    """Test when the root data passed in is a list instead of a dict."""
    table = Table()
    table.add_column("Key")
    table.add_column("Value")

    data = [{"root_list_item1": "valA"}, {"root_list_item2": "valB"}]

    _build_annotation_table("test_key", table, data, is_stub=False)
    output = _render_table_to_string(table)

    assert "root_list_item1:" in output
    assert "valA" in output
    assert "root_list_item2:" in output
    assert "valB" in output



# Copyright 2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import pathlib
from typing import Annotated, Optional, List

import typer

logger = logging.getLogger(__name__)


def parse_adapter(
    ctx: typer.Context,
    file_path: Annotated[
        pathlib.Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="Path to the source file to parse (e.g., a .srt or .txt file).",
        ),
    ],
    schema_id: Annotated[
        str,
        typer.Argument(
            help="The target Dorsal schema ID to parse into (e.g., 'open/audio-transcription').",
        ),
    ],
    source_format: Annotated[
        Optional[str],
        typer.Option(
            "--format",
            "-f",
            help="The format of the source file (e.g., 'srt', 'vtt'). Inferred from extension if omitted.",
        ),
    ] = None,
    options: Annotated[
        Optional[List[str]],
        typer.Option(
            "--opt",
            "-o",
            help="Runtime options for the parser in 'key=value' format. Can be used multiple times.",
        ),
    ] = None,
    output_path: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            "--output",
            "-out",
            help="Custom output file/directory path. Overrides auto-save naming.",
        ),
    ] = None,
    no_save: Annotated[
        bool,
        typer.Option(
            "--no-save",
            help="Disable auto-saving results to disk (useful for pure piping to stdout).",
        ),
    ] = False,
):
    """
    Parse a standard file format into a valid Dorsal JSON record using an adapter.
    """
    from dorsal.common.cli import exit_cli, EXIT_CODE_ERROR, get_rich_console, get_error_console, parse_cli_options
    from dorsal.common.exceptions import DorsalError
    from dorsal.api.adapters import parse_file_from_path

    console = get_rich_console()
    error_console = get_error_console()
    palette: dict[str, str] = ctx.obj.get("palette", {}) if ctx.obj else {}

    parsed_options = parse_cli_options(options=options, palette=palette)

    try:
        record_dict = parse_file_from_path(
            file_path=file_path, schema_id=schema_id, source_format=source_format, **parsed_options
        )
    except DorsalError as e:
        error_console.print(f"[{palette.get('error', 'bold red')}]Parse Failed:[/] {e}")
        exit_cli(code=EXIT_CODE_ERROR)
    except Exception as e:
        logger.exception("Unexpected error during parsing.")
        error_console.print(f"[{palette.get('error', 'bold red')}]Unexpected Error:[/] {e}")
        exit_cli(code=EXIT_CODE_ERROR)

    final_payload = {
        "schema_id": schema_id,
        "file_path": str(file_path.resolve()),
        "record": record_dict,
    }

    out_dir = output_path if output_path and output_path.is_dir() else pathlib.Path.cwd()
    save_path = output_path if output_path and not output_path.is_dir() else out_dir / f"{file_path.stem}.dorsal.json"

    saved_path_msg = None
    if not no_save:
        try:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            save_path.write_text(json.dumps(final_payload, indent=2, ensure_ascii=False), encoding="utf-8")
            saved_path_msg = str(save_path.resolve())
        except Exception as e:
            error_console.print(f"\n[{palette.get('error', 'bold red')}]Failed to write output file:[/] {e}")
            exit_cli(code=EXIT_CODE_ERROR)

    console.print(json.dumps(final_payload, indent=2, ensure_ascii=False))

    if not no_save and saved_path_msg:
        error_console.print(
            f"\n[{palette.get('info', 'dim')}]Parsed record saved successfully:\n  ↳ {saved_path_msg}[/]"
        )



# Copyright 2025-2026 Dorsal Hub LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import datetime
import itertools
import logging
import uuid
from typing import Annotated, Any, Callable, Literal, Self, Type, Union

from pydantic import (
    AnyHttpUrl,
    AwareDatetime,
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    NonNegativeInt,
    ValidationInfo,
    field_validator,
    model_validator,
)

import dorsal
from dorsal.common.constants import ANNOTATION_MAX_SIZE_BYTES, ANNOTATION_SCHEMA_LIMIT_STRICT
from dorsal.common.exceptions import PydanticValidationError
from dorsal.common.model import (
    AnnotationSource,
)
from dorsal.file.utils.size import check_record_size
from dorsal.common.validators import DatasetID, Pagination, String256, TString255
from dorsal.file.validators.base import (
    FileCoreValidationModel,
    FileExtension,
    FILESIZE_UPPER_LIMIT,
    MediaTypeString,
)
from dorsal.file.validators.common import Blake3Hash, QuickHash, SHA256Hash, TLSHash
from dorsal.file.validators.mediainfo import MediaInfoValidationModel
from dorsal.file.validators.pdf import PDFValidationModel
from dorsal.file.validators.ebook import EbookValidationModel
from dorsal.file.validators.office_document import OfficeDocumentValidationModel


logger = logging.getLogger(__name__)


class GenericFileAnnotation(BaseModel):
    """A container for arbitrary annotation record data, allowing any fields.

    Used in `LocalFile._add_annotation` when the annotation being added is a dictionary.

    """

    model_config = ConfigDict(extra="allow")


class AnnotationGroupInfo(BaseModel):
    id: uuid.UUID = Field(description="ID linking all chunks of a split annotation.")
    index: int = Field(ge=0)
    total: int = Field(gt=1, description="Total number of chunks in the entire annotation.")


class Annotation(BaseModel):
    """Container for a single annotation record (e.g. `file/pdf`)

    Attributes:
    - `record` - e.g. `PDFValidationModel` - a single file annotation record
    - `source` - has attributes (e.g. `version`) relevant to production of `record`
    """

    record: GenericFileAnnotation | None
    private: bool | None = None
    source: AnnotationSource
    schema_id: str | None = None
    schema_version: str | None = None
    group: AnnotationGroupInfo | None = None

    @field_validator("record", mode="after")
    @classmethod
    def _validate_record_size(cls, record: GenericFileAnnotation | None) -> GenericFileAnnotation | None:
        """Enforces global annotation size limit."""
        if record is None:
            return None

        payload_size = check_record_size(record)

        if payload_size > ANNOTATION_MAX_SIZE_BYTES:
            raise ValueError(
                f"Annotation record size ({payload_size} bytes) exceeds the limit of "
                f"{ANNOTATION_MAX_SIZE_BYTES} bytes. Please ensure sharding is applied."
            )

        return record


class AnnotationXL(Annotation):
    """An Annotation variant that bypasses the size limit check.

    Used by `LocalFile.get_annotations`
    """

    @field_validator("record", mode="after")
    @classmethod
    def _validate_record_size(cls, record: GenericFileAnnotation | None) -> GenericFileAnnotation | None:
        return record


class ShardedAnnotation(Annotation):
    """A strict subset of Annotation that guarantees 'record' and 'group' are present."""

    record: GenericFileAnnotation
    group: AnnotationGroupInfo


class AnnotationGroup(BaseModel):
    """Container for multiple linked annotations. The order in the array is meaningful."""

    annotations: list[Annotation] = Field(min_length=1, max_length=100)

    @model_validator(mode="after")
    def _similarity_check(self) -> Self:
        private = self.annotations[0].private
        source = self.annotations[0].source
        schema_version = self.annotations[0].schema_version
        for ann in self.annotations[1:]:
            if ann.private != private:
                raise ValueError("Field `private` must be identical across all Annotation records")
            if ann.source != source:
                raise ValueError("Field `source` must be identical across all Annotation records")
            if ann.schema_version != schema_version:
                raise ValueError("Field `schema_version` must be identical across all Annotation records")

        return self

    @property
    def private(self) -> bool | None:
        return self.annotations[0].private

    @property
    def source(self) -> AnnotationSource:
        return self.annotations[0].source

    @property
    def schema_version(self) -> str | None:
        return self.annotations[0].schema_version


class Annotation_Base(Annotation):
    record: FileCoreValidationModel  # type: ignore[assignment]
    private: bool | None = Field(default=None, exclude=True)


class Annotation_OfficeDocument(Annotation):
    record: OfficeDocumentValidationModel | None = None  # type: ignore[assignment]
    private: bool | None = None


class Annotation_MediaInfo(Annotation):
    record: MediaInfoValidationModel | None = None  # type: ignore[assignment]
    private: bool | None = None


class Annotation_PDF(Annotation):
    record: PDFValidationModel | None = None  # type: ignore[assignment]
    private: bool | None = None


class Annotation_Ebook(Annotation):
    record: EbookValidationModel | None = None  # type: ignore[assignment]
    private: bool | None = None


CORE_MODEL_ANNOTATION_WRAPPERS: dict[str | None, Type[Annotation]] = {
    "file/base": Annotation_Base,
    "file/ebook": Annotation_Ebook,
    "file/office": Annotation_OfficeDocument,
    "file/mediainfo": Annotation_MediaInfo,
    "file/pdf": Annotation_PDF,
}


class AnnotationStub(BaseModel):
    """
    Represents a lightweight, summary view of an annotation,
    returned in a fully hydrated FileRecord.
    """

    model_config = ConfigDict(populate_by_name=True)

    hash: SHA256Hash
    id: uuid.UUID
    source: AnnotationSource
    user_id: int = Field(validation_alias=AliasChoices("user_no", "user_id"))
    date_modified: AwareDatetime
    group: AnnotationGroupInfo | None = None


class Annotations(BaseModel):
    """
    File annotation container.
    """

    model_config = ConfigDict(
        extra="allow",
        populate_by_name=True,
    )

    file_base: Annotation_Base = Field(alias="file/base")
    file_ebook: Annotation_Ebook | None = Field(default=None, alias="file/ebook")
    file_mediainfo: Annotation_MediaInfo | None = Field(default=None, alias="file/mediainfo")
    file_office: Annotation_OfficeDocument | None = Field(default=None, alias="file/office")
    file_pdf: Annotation_PDF | None = Field(default=None, alias="file/pdf")

    @model_validator(mode="wrap")
    @classmethod
    def _validate_and_type_extras(cls, value: Any, handler: Callable[[Any], Self], info: ValidationInfo) -> Self:
        validated_instance = handler(value)
        current_extras = validated_instance.__pydantic_extra__

        if current_extras:
            for key in list(current_extras.keys()):
                extra_value = current_extras[key]
                parsed_value: Any = None

                try:
                    if isinstance(extra_value, dict):
                        if "annotations" in extra_value and isinstance(extra_value["annotations"], list):
                            parsed_value = AnnotationGroup.model_validate(extra_value)
                        elif "record" in extra_value:
                            parsed_value = Annotation.model_validate(extra_value)
                        else:
                            parsed_value = AnnotationStub.model_validate(extra_value)

                    elif isinstance(extra_value, list) and extra_value:
                        new_list: list[Annotation | AnnotationGroup | AnnotationStub] = []
                        for item in extra_value:
                            if isinstance(item, dict):
                                if "annotations" in item:
                                    new_list.append(AnnotationGroup.model_validate(item))

                                elif "record" in item:
                                    new_list.append(Annotation.model_validate(item))

                                else:
                                    new_list.append(AnnotationStub.model_validate(item))

                            elif isinstance(item, (Annotation, AnnotationGroup, AnnotationStub)):
                                new_list.append(item)
                            else:
                                raise ValueError(f"List item is not a valid Annotation type: {type(item)}")

                        parsed_value = new_list

                    else:
                        parsed_value = extra_value

                    setattr(validated_instance, key, parsed_value)
                    validated_instance.model_fields_set.add(key)

                except PydanticValidationError as err:
                    error_details = err.errors(include_url=False)
                    raise ValueError(
                        f"Annotation key '{key}' contains invalid data. "
                        f"Could not validate as Annotation, Group, or Stub. "
                        f"Details: {error_details}"
                    ) from err
                except Exception as err:
                    raise ValueError(
                        f"An unexpected error occurred while validating annotation key '{key}': {str(err)}"
                    ) from err
        return validated_instance


class AnnotationsStrict(Annotations):
    """
    Strict version of Annotations for write/submission operations.

    Enforces:
    1. Structure Limits: Max 64 datasets, Max 64 entries per dataset.
    2. Data Completeness: No AnnotationStubs allowed. Must be full Annotation or Group.
    """

    @model_validator(mode="before")
    @classmethod
    def _validate_submission_limits(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value

        if len(value) > 64:  # Allows for every open validation schema with 54 to spare for custom schemas
            raise ValueError(f"Too many annotation schemas. Received {len(value)}, limit is 64 schemas per file.")

        for key, items in value.items():
            if isinstance(items, list) and len(items) > ANNOTATION_SCHEMA_LIMIT_STRICT:
                raise ValueError(
                    f"Too many '{key}' annotations. {len(items)} exceeds the limit ({ANNOTATION_SCHEMA_LIMIT_STRICT})."
                )
        return value

    @model_validator(mode="after")
    def _validate_no_stubs(self) -> Self:
        if not self.__pydantic_extra__:
            return self

        for key, value in self.__pydantic_extra__.items():
            items = value if isinstance(value, list) else [value]

            for item in items:
                if isinstance(item, AnnotationStub):
                    raise ValueError(
                        f"Invalid data for '{key}': Contains one or more `AnnotationStub` objects. "
                        "`FileRecordStrict` cannot contain `AnnotationStub`."
                    )
        return self


class FileTag(BaseModel):
    id: str | None = Field(pattern=r"^[0-9a-f]{24}$", default=None)  # bson.ObjectId string
    name: str = Field(pattern=r"^[a-zA-Z0-9\_]{3,64}$")
    value: String256 | bool | datetime.datetime | int | float
    value_code: String256 | None = None
    private: bool | None = None
    hidden: bool
    upvotes: NonNegativeInt
    downvotes: NonNegativeInt
    origin: Literal["dorsal.LocalFile", "DorsalHub"] = "DorsalHub"


class NewFileTag(FileTag):
    hidden: Literal[False] = False
    upvotes: Literal[0] = 0
    downvotes: Literal[0] = 0
    origin: Literal["dorsal.LocalFile"] = "dorsal.LocalFile"


class ValidTag(BaseModel):
    name: str
    value: Any


class ValidateTagsResult(BaseModel):
    valid: bool
    message: str | None = None
    tags: list[ValidTag] = Field(default_factory=list)


class TagResult(BaseModel):
    name: str
    namespace: str
    value: Any
    value_code: Any | None = None
    private: bool
    status: Literal["added", "no-action", "upvoted"]
    detail: str | None = None


class UrlSource(BaseModel):
    """
    Tracks the provenance and discovery context of a File URL.
    """

    agent: str = Field(default="dorsal-python", max_length=64, description="The software/tool that added this URL.")
    agent_version: str | None = Field(default=dorsal.__version__, max_length=32)
    parent_url: HttpUrl | None = Field(default=None, description="The webpage or directory where this link was found.")
    reference: str | None = Field(default=None, max_length=128, description="Context string or detail about the URL.")
    user_id: int | None = Field(
        default=None,
        validation_alias=AliasChoices("user_no", "user_id"),
        description="ID of the user who added the URL",
    )


class FileUrl(BaseModel):
    """File URL record."""

    id: str | None = Field(pattern=r"^url_[a-zA-Z0-9_-]{22}$", default=None)
    url: AnyHttpUrl
    private: bool | None = None
    source: UrlSource
    status: Literal["alive", "dead", "restricted", "unreachable", "unverified"]
    first_added: AwareDatetime | None = None
    last_checked: AwareDatetime | None = None
    upvotes: NonNegativeInt = 0
    downvotes: NonNegativeInt = 0


class NewFileUrl(FileUrl):
    status: Literal["unverified"] = "unverified"


DeletionScope = Literal["all", "public", "private", "none"]


class FileRecord(BaseModel):
    """File annotation record.

    Optionally includes validation hash.

    """

    hash: SHA256Hash | None = None
    validation_hash: Blake3Hash | None = None
    quick_hash: QuickHash | None = None
    similarity_hash: TLSHash | None = None
    annotations: Annotations | None = None
    tags: list[FileTag] = Field(default_factory=list)
    urls: list[FileUrl] = Field(default_factory=list)

    def _validate_or_populate_hash_field(
        self, *, field_name: str, source_value: str | None, source_name_for_error: str
    ):
        if source_value is None:
            return

        current_value = getattr(self, field_name)
        if current_value is None:
            setattr(self, field_name, source_value)
        elif current_value != source_value:
            raise ValueError(
                f"Data integrity error: Top-level '{field_name}' ('{current_value}') "
                f"does not match the source '{source_name_for_error}' value "
                f"('{source_value}') from 'file/base'."
            )

    @model_validator(mode="after")
    def _populate_and_validate_hashes_from_base(self) -> Self:
        """
        - Populates the top-level `quick_hash`, `similarity_hash` and `validation_hash` fields, when available in `file/base`.
        - Validates and enforces hash consistency between the top-level hash fields and 'file/base'.
        """
        if not self.annotations or not self.annotations.file_base:
            return self

        core_validation_model = self.annotations.file_base.record

        if self.hash != core_validation_model.hash:
            raise ValueError(
                "Data integrity error: Top-level 'hash' does not match the 'hash' "
                f"in the 'file/base' annotation record ('{self.hash}' vs '{core_validation_model.hash}')."
            )

        self._validate_or_populate_hash_field(
            field_name="quick_hash",
            source_value=core_validation_model.quick_hash,
            source_name_for_error="QUICK",
        )
        self._validate_or_populate_hash_field(
            field_name="similarity_hash",
            source_value=core_validation_model.similarity_hash,
            source_name_for_error="TLSH",
        )
        self._validate_or_populate_hash_field(
            field_name="validation_hash",
            source_value=(
                core_validation_model.all_hash_ids.get("BLAKE3") if core_validation_model.all_hash_ids else None
            ),
            source_name_for_error="BLAKE3",
        )
        return self

    @model_validator(mode="after")
    def _identical_hash_check(self) -> Self:
        """
        Ensures no two identical hashes are populated.
        """
        hash_fields = [
            self.hash,
            self.validation_hash,
            self.quick_hash,
            self.similarity_hash,
        ]
        hashes = [h for h in hash_fields if h is not None]

        if len(set(hashes)) == len(hashes):
            return self

        field_names = ["hash", "validation_hash", "quick_hash", "similarity_hash"]
        for field1, field2 in itertools.combinations(field_names, 2):
            hash1 = getattr(self, field1, None)
            hash2 = getattr(self, field2, None)

            if hash1 is not None and hash1 == hash2:
                raise ValueError(
                    f"Inconsistent hash values: '{field1}' and '{field2}' are "
                    f"unexpectedly identical (value: '{hash1}')."
                )

        raise ValueError("Inconsistent hash values: A duplicate was detected but could not be located.")


class FileRecordStrict(FileRecord):
    """File annotation record, including validation hash.

    - Used for indexing new files and their metadata (including structured records and/or tags)
    - This format is generated locally by a ModelRunner instance, it is not provided by the DorsalHub API.

    Used in:

    - `POST /file/public|private`
    """

    hash: SHA256Hash
    validation_hash: Blake3Hash
    annotations: AnnotationsStrict
    source: Literal["disk", "cache", "dorsalhub"]

    tags: list[FileTag] = Field(default_factory=list, max_length=128)
    urls: list[FileUrl] = Field(default_factory=list, max_length=10)


class FileRecordDateTime(FileRecord):
    """A `FileRecord` with dates in the root level."""

    hash: SHA256Hash
    all_hashes: None = Field(default=None, exclude=True)
    all_hash_ids: None = Field(default=None, exclude=True)
    date_created: AwareDatetime
    date_modified: AwareDatetime


class FileRecordLite(BaseModel):
    """Flat file record, with no secure hashes."""

    quick_hash: SHA256Hash
    name: TString255
    extension: FileExtension | None = None
    size: int = Field(ge=0, lt=FILESIZE_UPPER_LIMIT)
    media_type: MediaTypeString

    annotations: Annotations | None = None
    tags: list[FileTag] = Field(default_factory=list)
    urls: list[FileUrl] = Field(default_factory=list)


class FileRecordSuperLite(BaseModel):
    """Flat file record, with no hashes.

    Used in cli/stats

    Note: for compatability with the rest of the pipeline, uuid4 is generated on the fly
    """

    hash: str
    name: TString255
    extension: FileExtension | None = None
    size: int = Field(ge=0, lt=FILESIZE_UPPER_LIMIT)
    media_type: MediaTypeString

    annotations: Annotations | None = None
    tags: list[FileTag] = Field(default_factory=list)
    urls: list[FileUrl] = Field(default_factory=list)


class AnnotationData(BaseModel):
    schema_id: DatasetID
    annotation: Annotation | None
    error: str | None = None


class UserFileRead(BaseModel):
    """User file record, for API response."""

    hash_: str = Field(alias="hash")
    user_id: int
    name: str
    extension: str | None
    size: int
    media_type: str
    private: bool
    first_indexed: datetime.datetime
    last_indexed: datetime.datetime


class FileSearchResponse(BaseModel):
    results: list[FileRecordDateTime]
    errors: list[str]
    pagination: Pagination
    api_version: str
